# Generated by Django 4.0.10 on 2023-07-11 12:53

import django.core.validators
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("memberaudit", "0010_alter_characterupdatestatus_section_and_more"),
    ]

    operations = [
        migrations.AlterField(
            model_name="characterskill",
            name="active_skill_level",
            field=models.PositiveIntegerField(
                validators=[
                    django.core.validators.MinValueValidator(0),
                    django.core.validators.MaxValueValidator(5),
                ]
            ),
        ),
        migrations.AlterField(
            model_name="characterskill",
            name="trained_skill_level",
            field=models.PositiveBigIntegerField(
                validators=[
                    django.core.validators.MinValueValidator(0),
                    django.core.validators.MaxValueValidator(5),
                ]
            ),
        ),
        migrations.AlterField(
            model_name="characterupdatestatus",
            name="section",
            field=models.CharField(
                choices=[
                    ("assets", "assets"),
                    ("attributes", "attributes"),
                    ("character_details", "character details"),
                    ("contacts", "contacts"),
                    ("contracts", "contracts"),
                    ("corporation_history", "corporation history"),
                    ("fw_stats", "faction warfare statistics"),
                    ("implants", "implants"),
                    ("jump_clones", "jump clones"),
                    ("location", "location"),
                    ("loyalty", "loyalty"),
                    ("mails", "mails"),
                    ("mining_ledger", "mining ledger"),
                    ("online_status", "online status"),
                    ("planets", "planets"),
                    ("ship", "ship"),
                    ("skills", "skills"),
                    ("skill_queue", "skill queue"),
                    ("skill_sets", "skill sets"),
                    ("standings", "standings"),
                    ("wallet_balance", "wallet balance"),
                    ("wallet_journal", "wallet journal"),
                    ("wallet_transactions", "wallet transactions"),
                ],
                db_index=True,
                max_length=64,
            ),
        ),
        migrations.AddConstraint(
            model_name="characterstanding",
            constraint=models.UniqueConstraint(
                fields=("character", "eve_entity"),
                name="functional_pk_characterstanding",
            ),
        ),
    ]



from unittest.mock import patch

from django.test import TestCase

from allianceauth.eveonline.models import EveAllianceInfo, EveCharacter
from allianceauth.tests.auth_utils import AuthUtils

from memberaudit.models import Character
from memberaudit.tests.testdata.factories import (
    create_character,
    create_character_from_user,
    create_character_update_status,
)
from memberaudit.tests.testdata.load_entities import load_entities
from memberaudit.tests.utils import (
    add_auth_character_to_user,
    add_memberaudit_character_to_user,
    create_memberaudit_character,
    create_user_from_evecharacter_with_access,
)

MODELS_PATH = "memberaudit.models.characters"


class TestCharacterQuerySet(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_entities()

    def test_should_return_set_of_eve_character_ids(self):
        # given
        create_memberaudit_character(1001)
        create_memberaudit_character(1002)
        # when/then
        self.assertSetEqual(Character.objects.all().eve_character_ids(), {1001, 1002})

    def test_should_return_characters_owner_by_user_only(self):
        # given
        character_1001 = create_memberaudit_character(1001)
        user = character_1001.character_ownership.user
        create_memberaudit_character(1002)
        # when
        result = Character.objects.owned_by_user(user)
        # then
        character_ids = {
            obj.character_ownership.character.character_id for obj in result
        }
        self.assertSetEqual(character_ids, {1001})

    def test_should_return_no_characters(self):
        # given
        user = AuthUtils.create_user("dummy")
        create_memberaudit_character(1001)
        # when
        result = Character.objects.owned_by_user(user)
        # then
        character_ids = {
            obj.character_ownership.character.character_id for obj in result
        }
        self.assertSetEqual(character_ids, set())


# Includes testing of Character.calc_total_update_status() to ensure they are in sync
class TestCharacterAnnotateTotalUpdateStatus(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_entities()
        cls.user, _ = create_user_from_evecharacter_with_access(1001)

    @patch(MODELS_PATH + ".MEMBERAUDIT_FEATURE_ROLES_ENABLED", True)
    def test_should_annotate_ok(self):
        # given
        character = create_character_from_user(self.user)
        for section in Character.UpdateSection:
            create_character_update_status(character, section=section)
        # when/then
        self.assertEqual(
            character.calc_total_update_status(), Character.TotalUpdateStatus.OK
        )
        # when
        qs = Character.objects.annotate_total_update_status()
        # then
        obj = qs.first()
        self.assertEqual(obj.total_update_status, Character.TotalUpdateStatus.OK)

    @patch(MODELS_PATH + ".MEMBERAUDIT_FEATURE_ROLES_ENABLED", False)
    def test_should_annotate_ok_when_all_enabled_sections_are_ok(self):
        # given
        character = create_character_from_user(self.user)
        for section in Character.UpdateSection.enabled_sections():
            create_character_update_status(character, section=section)
        create_character_update_status(
            character=character, is_success=False, section=Character.UpdateSection.ROLES
        )
        # when/then
        self.assertEqual(
            character.calc_total_update_status(), Character.TotalUpdateStatus.OK
        )
        # when
        qs = Character.objects.annotate_total_update_status()
        # then
        obj = qs.first()
        self.assertEqual(obj.total_update_status, Character.TotalUpdateStatus.OK)

    @patch(MODELS_PATH + ".MEMBERAUDIT_FEATURE_ROLES_ENABLED", True)
    def test_should_annotate_error(self):
        # given
        character = create_character_from_user(self.user)
        create_character_update_status(
            character, section=Character.UpdateSection.ASSETS, is_success=False
        )
        # when/then
        self.assertEqual(
            character.calc_total_update_status(), Character.TotalUpdateStatus.ERROR
        )
        # when
        qs = Character.objects.annotate_total_update_status()
        # then
        obj = qs.first()
        self.assertEqual(obj.total_update_status, Character.TotalUpdateStatus.ERROR)

    @patch(MODELS_PATH + ".MEMBERAUDIT_FEATURE_ROLES_ENABLED", True)
    def test_should_annotate_incomplete(self):
        # given
        character = create_character_from_user(self.user)
        sections_to_update = [
            obj
            for obj in Character.UpdateSection
            if obj != Character.UpdateSection.ASSETS
        ]
        for section in sections_to_update:
            create_character_update_status(character, section=section)
        # when/then
        self.assertEqual(
            character.calc_total_update_status(), Character.TotalUpdateStatus.INCOMPLETE
        )
        # when
        qs = Character.objects.annotate_total_update_status()
        # then
        obj = qs.first()
        self.assertEqual(
            obj.total_update_status, Character.TotalUpdateStatus.INCOMPLETE
        )

    @patch(MODELS_PATH + ".MEMBERAUDIT_FEATURE_ROLES_ENABLED", True)
    def test_should_annotate_in_progress(self):
        # given
        character = create_character_from_user(self.user)
        for section in Character.UpdateSection:
            if section == Character.UpdateSection.ASSETS:
                create_character_update_status(
                    character, section=section, is_success=None
                )
            else:
                create_character_update_status(character, section=section)
        # when/then
        self.assertEqual(
            character.calc_total_update_status(),
            Character.TotalUpdateStatus.IN_PROGRESS,
        )
        # when
        qs = Character.objects.annotate_total_update_status()
        # then
        obj = qs.first()
        self.assertEqual(
            obj.total_update_status, Character.TotalUpdateStatus.IN_PROGRESS
        )

    def test_should_annotate_disabled(self):
        # given
        character = create_character_from_user(self.user, is_disabled=True)
        # when/then
        self.assertEqual(
            character.calc_total_update_status(), Character.TotalUpdateStatus.DISABLED
        )
        # when
        qs = Character.objects.annotate_total_update_status()
        # then
        obj = qs.first()
        self.assertEqual(obj.total_update_status, Character.TotalUpdateStatus.DISABLED)

    @patch(MODELS_PATH + ".MEMBERAUDIT_FEATURE_ROLES_ENABLED", True)
    def test_should_annotate_limited_token_when_one_token_issue_only(self):
        # given
        character = create_character_from_user(self.user)
        create_character_update_status(
            character,
            section=Character.UpdateSection.ASSETS,
            is_success=False,
            has_token_error=True,
        )
        # when/then
        self.assertEqual(
            character.calc_total_update_status(),
            Character.TotalUpdateStatus.LIMITED_TOKEN,
        )
        # when
        qs = Character.objects.annotate_total_update_status()
        # then
        obj = qs.first()
        self.assertEqual(
            obj.total_update_status, Character.TotalUpdateStatus.LIMITED_TOKEN
        )

    @patch(MODELS_PATH + ".MEMBERAUDIT_FEATURE_ROLES_ENABLED", True)
    def test_should_annotate_error_when_several_token_issues(self):
        # given
        character = create_character_from_user(self.user)
        create_character_update_status(
            character,
            section=Character.UpdateSection.ASSETS,
            is_success=False,
            has_token_error=True,
        )
        create_character_update_status(
            character,
            section=Character.UpdateSection.LOCATION,
            is_success=False,
            has_token_error=True,
        )
        # when/then
        self.assertEqual(
            character.calc_total_update_status(), Character.TotalUpdateStatus.ERROR
        )
        # when
        qs = Character.objects.annotate_total_update_status()
        # then
        obj = qs.first()
        self.assertEqual(obj.total_update_status, Character.TotalUpdateStatus.ERROR)


class TestCharacterUserHasScope(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_entities()
        # main character with alts
        cls.character_1001 = create_memberaudit_character(1001)  # main
        cls.character_1110 = add_memberaudit_character_to_user(  # alt
            cls.character_1001.eve_character.character_ownership.user, 1110
        )
        cls.character_1121 = add_memberaudit_character_to_user(  # alt
            cls.character_1001.eve_character.character_ownership.user, 1121
        )
        # main character with alts
        cls.character_1002 = create_memberaudit_character(1002)
        cls.character_1002.is_shared = True
        cls.character_1002.save()
        cls.character_1103 = add_memberaudit_character_to_user(
            cls.character_1002.eve_character.character_ownership.user, 1103
        )
        # main characters
        cls.character_1003 = create_memberaudit_character(1003)
        cls.character_1101 = create_memberaudit_character(1101)
        cls.character_1102 = create_memberaudit_character(1102)
        cls.character_1102.is_shared = True
        cls.character_1102.save()
        cls.character_1111 = create_memberaudit_character(1111)
        cls.character_1122 = create_memberaudit_character(1122)
        cls.member_state = AuthUtils.get_member_state()
        cls.member_state.member_alliances.add(
            EveAllianceInfo.objects.get(alliance_id=3001)
        )

    def test_user_owning_character_has_scope(self):
        """
        when user is the owner of characters
        then include those characters only
        """
        result_qs = Character.objects.user_has_scope(
            user=self.character_1001.eve_character.character_ownership.user
        )
        self.assertSetEqual(result_qs.eve_character_ids(), {1001, 1110, 1121})

    def test_view_own_corporation_1(self):
        """
        when user has scope to view own corporation
        then include characters of corporations members (mains + alts)
        """
        user = self.character_1001.eve_character.character_ownership.user
        user = AuthUtils.add_permission_to_user_by_name(
            "memberaudit.view_same_corporation", user
        )
        result_qs = Character.objects.user_has_scope(user=user)
        self.assertSetEqual(
            result_qs.eve_character_ids(), {1001, 1110, 1121, 1002, 1103}
        )

    def test_view_own_alliance_1(self):
        """
        when user has scope to view own alliance
        then include characters of alliance members (mains + alts)
        """
        user = self.character_1001.eve_character.character_ownership.user
        user = AuthUtils.add_permission_to_user_by_name(
            "memberaudit.view_same_alliance", user
        )
        result_qs = Character.objects.user_has_scope(user=user)
        self.assertSetEqual(
            result_qs.eve_character_ids(), {1001, 1110, 1121, 1002, 1003, 1103}
        )

    def test_view_own_alliance_2(self):
        """
        when user has permission to view own alliance
        and does not belong to any alliance
        then do not include any alliance characters
        """
        user = self.character_1102.eve_character.character_ownership.user
        user = AuthUtils.add_permission_to_user_by_name(
            "memberaudit.view_same_alliance", user
        )
        result_qs = Character.objects.user_has_scope(user=user)
        self.assertSetEqual(result_qs.eve_character_ids(), {1102})

    def test_view_everything_1(self):
        """
        when user has scope to view everything
        then include all characters
        """
        user = self.character_1001.eve_character.character_ownership.user
        user = AuthUtils.add_permission_to_user_by_name(
            "memberaudit.view_everything", user
        )
        result_qs = Character.objects.user_has_scope(user=user)
        self.assertSetEqual(
            result_qs.eve_character_ids(),
            {1001, 1002, 1003, 1101, 1102, 1103, 1110, 1111, 1121, 1122},
        )


class TestCharacterUserHasAccess(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_entities()
        # main character with alts
        cls.character_1001 = create_memberaudit_character(1001)  # main
        cls.character_1110 = add_memberaudit_character_to_user(  # alt
            cls.character_1001.eve_character.character_ownership.user, 1110
        )
        cls.character_1121 = add_memberaudit_character_to_user(  # alt
            cls.character_1001.eve_character.character_ownership.user, 1121
        )
        # main character with alts
        cls.character_1002 = create_memberaudit_character(1002)
        cls.character_1002.is_shared = True
        cls.character_1002.save()
        AuthUtils.add_permission_to_user_by_name(
            "memberaudit.share_characters",
            cls.character_1002.eve_character.character_ownership.user,
        )
        cls.character_1103 = add_memberaudit_character_to_user(
            cls.character_1002.eve_character.character_ownership.user, 1103
        )
        # main characters
        cls.character_1003 = create_memberaudit_character(1003)
        cls.character_1101 = create_memberaudit_character(1101)
        cls.character_1102 = create_memberaudit_character(1102)
        cls.character_1102.is_shared = True
        cls.character_1102.save()
        AuthUtils.add_permission_to_user_by_name(
            "memberaudit.share_characters",
            cls.character_1102.eve_character.character_ownership.user,
        )
        cls.character_1111 = create_memberaudit_character(1111)
        cls.character_1122 = create_memberaudit_character(1122)
        cls.member_state = AuthUtils.get_member_state()
        cls.member_state.member_alliances.add(
            EveAllianceInfo.objects.get(alliance_id=3001)
        )

    def test_user_owning_character_has_access(self):
        """
        when user is the owner of characters
        then include those characters only
        """
        result_qs = Character.objects.user_has_access(
            user=self.character_1001.eve_character.character_ownership.user
        )
        self.assertSetEqual(result_qs.eve_character_ids(), {1001, 1110, 1121})

    def test_view_own_corporation_1(self):
        """
        when user has permission to view own corporation and not characters_access
        then include own characters only
        """
        user = self.character_1001.eve_character.character_ownership.user
        user = AuthUtils.add_permission_to_user_by_name(
            "memberaudit.view_same_corporation", user
        )
        result_qs = Character.objects.user_has_access(user=user)
        self.assertSetEqual(result_qs.eve_character_ids(), {1001, 1110, 1121})

    def test_view_own_corporation_2(self):
        """
        when user has permission to view own corporation and characters_access
        then include characters of corporations members (mains + alts)
        """
        user = self.character_1001.eve_character.character_ownership.user
        user = AuthUtils.add_permission_to_user_by_name(
            "memberaudit.view_same_corporation", user
        )
        user = AuthUtils.add_permission_to_user_by_name(
            "memberaudit.characters_access", user
        )
        result_qs = Character.objects.user_has_access(user=user)
        self.assertSetEqual(
            result_qs.eve_character_ids(), {1001, 1110, 1121, 1002, 1103}
        )

    def test_view_own_alliance_1a(self):
        """
        when user has permission to view own alliance and not characters_access
        then include own character only
        """
        user = self.character_1001.eve_character.character_ownership.user
        user = AuthUtils.add_permission_to_user_by_name(
            "memberaudit.view_same_alliance", user
        )
        result_qs = Character.objects.user_has_access(user=user)
        self.assertSetEqual(result_qs.eve_character_ids(), {1001, 1110, 1121})

    def test_view_own_alliance_1b(self):
        """
        when user has permission to view own alliance and characters_access
        then include characters of alliance members (mains + alts)
        """
        user = self.character_1001.eve_character.character_ownership.user
        user = AuthUtils.add_permission_to_user_by_name(
            "memberaudit.view_same_alliance", user
        )
        user = AuthUtils.add_permission_to_user_by_name(
            "memberaudit.characters_access", user
        )
        result_qs = Character.objects.user_has_access(user=user)
        self.assertSetEqual(
            result_qs.eve_character_ids(), {1001, 1110, 1121, 1002, 1003, 1103}
        )

    def test_view_own_alliance_2(self):
        """
        when user has permission to view own alliance and characters_access
        and does not belong to any alliance
        then do not include any alliance characters
        """
        user = self.character_1102.eve_character.character_ownership.user
        user = AuthUtils.add_permission_to_user_by_name(
            "memberaudit.view_same_alliance", user
        )
        user = AuthUtils.add_permission_to_user_by_name(
            "memberaudit.characters_access", user
        )
        result_qs = Character.objects.user_has_access(user=user)
        self.assertSetEqual(result_qs.eve_character_ids(), {1102})

    def test_view_everything_1(self):
        """
        when user has permission to view everything and no characters_access
        then include own character only
        """
        user = self.character_1001.eve_character.character_ownership.user
        user = AuthUtils.add_permission_to_user_by_name(
            "memberaudit.view_everything", user
        )
        result_qs = Character.objects.user_has_access(user=user)
        self.assertSetEqual(result_qs.eve_character_ids(), {1001, 1110, 1121})

    def test_view_everything_2(self):
        """
        when user has permission to view everything and characters_access
        then include all characters
        """
        user = self.character_1001.eve_character.character_ownership.user
        user = AuthUtils.add_permission_to_user_by_name(
            "memberaudit.view_everything", user
        )
        user = AuthUtils.add_permission_to_user_by_name(
            "memberaudit.characters_access", user
        )
        result_qs = Character.objects.user_has_access(user=user)
        self.assertSetEqual(
            result_qs.eve_character_ids(),
            {1001, 1002, 1003, 1101, 1102, 1103, 1110, 1111, 1121, 1122},
        )

    def test_recruiter_access(self):
        """
        when user has recruiter permission
        then include own character plus shared characters from members
        """
        user = self.character_1001.eve_character.character_ownership.user
        user = AuthUtils.add_permission_to_user_by_name(
            "memberaudit.view_shared_characters", user
        )
        result_qs = Character.objects.user_has_access(user=user)
        self.assertSetEqual(
            result_qs.eve_character_ids(), {1001, 1002, 1102, 1110, 1121}
        )

    def test_recruiter_should_loose_access_once_recruit_becomes_member(self):
        # given
        character_1107 = create_memberaudit_character(1107)
        character_1107.is_shared = True
        character_1107.save()
        user = self.character_1001.eve_character.character_ownership.user
        user = AuthUtils.add_permission_to_user_by_name(
            "memberaudit.view_shared_characters", user
        )
        # when
        result_qs = Character.objects.user_has_access(user=user)
        self.assertNotIn(1107, result_qs.eve_character_ids())


class TestCharacterUnregisteredCharacterCount(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_entities()
        # main character with alts
        cls.character_1001 = create_memberaudit_character(1001)
        cls.user = cls.character_1001.character_ownership.user

    def test_should_return_zero_when_no_unregistered(self):
        # when
        result = Character.objects.characters_of_user_to_register_count(self.user)
        # then
        self.assertEqual(result, 0)

    def test_should_return_count_including_unregistered(self):
        # given
        add_auth_character_to_user(self.user, 1002)
        # when
        result = Character.objects.characters_of_user_to_register_count(self.user)
        # then
        self.assertEqual(result, 1)

    @patch(MODELS_PATH + ".MEMBERAUDIT_FEATURE_ROLES_ENABLED", True)
    def test_should_return_count_including_registered_with_token_error(self):
        # given
        character_1002 = add_memberaudit_character_to_user(self.user, 1002)
        create_character_update_status(
            character_1002,
            section=Character.UpdateSection.ASSETS,
            is_success=False,
            has_token_error=True,
            error_message="TokenError 1",
        )
        create_character_update_status(
            character_1002,
            section=Character.UpdateSection.CONTRACTS,
            is_success=False,
            error_message="TokenError 2",
        )
        # when
        result = Character.objects.characters_of_user_to_register_count(self.user)
        # then
        self.assertEqual(result, 1)

    @patch(MODELS_PATH + ".MEMBERAUDIT_FEATURE_ROLES_ENABLED", False)
    def test_should_return_count_not_including_token_errors_for_disabled_sections(self):
        # given
        character_1002 = add_memberaudit_character_to_user(self.user, 1002)
        create_character_update_status(
            character_1002,
            section=Character.UpdateSection.ROLES,
            is_success=False,
            has_token_error=True,
            error_message="TokenError 1",
        )
        # when
        result = Character.objects.characters_of_user_to_register_count(self.user)
        # then
        self.assertEqual(result, 0)

    @patch(MODELS_PATH + ".MEMBERAUDIT_FEATURE_ROLES_ENABLED", False)
    def test_should_return_count_disabled_characters(self):
        # given
        character_1002 = add_memberaudit_character_to_user(
            self.user, 1002, is_disabled=True
        )
        create_character_update_status(
            character_1002,
            section=Character.UpdateSection.ASSETS,
            is_success=False,
            has_token_error=True,
            error_message="TokenError 1",
        )

        # when
        result = Character.objects.characters_of_user_to_register_count(self.user)
        # then
        self.assertEqual(result, 1)

    @patch(MODELS_PATH + ".MEMBERAUDIT_FEATURE_ROLES_ENABLED", False)
    def test_should_not_count_disabled_and_token_errors_twice(self):
        # given
        add_memberaudit_character_to_user(self.user, 1002, is_disabled=True)
        # when
        result = Character.objects.characters_of_user_to_register_count(self.user)
        # then
        self.assertEqual(result, 1)


class TestCharacterDisableCharacterWithoutOwner(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_entities()
        cls.character = create_memberaudit_character(1001)

    def test_should_disable_orphans(self):
        # given
        orphan_1 = create_character(
            EveCharacter.objects.get(character_id=1121), is_disabled=False
        )
        orphan_2 = create_character(
            EveCharacter.objects.get(character_id=1111), is_disabled=False
        )
        # when
        result = Character.objects.disable_characters_with_no_owner()
        # then
        self.assertEqual(result, 2)
        orphan_1.refresh_from_db()
        self.assertTrue(orphan_1.is_disabled)
        orphan_2.refresh_from_db()
        self.assertTrue(orphan_2.is_disabled)
        self.assertFalse(self.character.is_disabled)

    def test_should_ignore_already_disables_orphans(self):
        # given
        orphan_disabled = create_character(
            EveCharacter.objects.get(character_id=1121), is_disabled=True
        )
        orphan_enabled = create_character(
            EveCharacter.objects.get(character_id=1111), is_disabled=False
        )
        # when
        result = Character.objects.disable_characters_with_no_owner()
        # then
        self.assertEqual(result, 1)
        orphan_disabled.refresh_from_db()
        self.assertTrue(orphan_disabled.is_disabled)
        orphan_enabled.refresh_from_db()
        self.assertTrue(orphan_enabled.is_disabled)
        self.assertFalse(self.character.is_disabled)

    def test_should_return_zero_when_nothing_to_disable(self):
        # given
        orphan_disabled = create_character(
            EveCharacter.objects.get(character_id=1121), is_disabled=True
        )
        # when
        result = Character.objects.disable_characters_with_no_owner()
        # then
        self.assertEqual(result, 0)
        orphan_disabled.refresh_from_db()
        self.assertTrue(orphan_disabled.is_disabled)
        self.assertFalse(self.character.is_disabled)


class TestCharacterUpdateStatusFilterEnabledSections(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_entities()
        cls.character_1001 = create_memberaudit_character(1001)

    @patch(MODELS_PATH + ".MEMBERAUDIT_FEATURE_ROLES_ENABLED", True)
    def test_should_return_enabled_sections_only_1(self):
        # given
        create_character_update_status(
            self.character_1001, section=Character.UpdateSection.ASSETS
        )
        create_character_update_status(
            self.character_1001, section=Character.UpdateSection.ROLES
        )
        # when
        result = self.character_1001.update_status_set.filter_enabled_sections()
        # then
        expected = {Character.UpdateSection.ASSETS, Character.UpdateSection.ROLES}
        sections = set(result.values_list("section", flat=True))
        self.assertSetEqual(sections, expected)

    @patch(MODELS_PATH + ".MEMBERAUDIT_FEATURE_ROLES_ENABLED", False)
    def test_should_return_enabled_sections_only_2(self):
        # given
        create_character_update_status(
            self.character_1001, section=Character.UpdateSection.ASSETS
        )
        create_character_update_status(
            self.character_1001, section=Character.UpdateSection.ROLES
        )
        # when
        result = self.character_1001.update_status_set.filter_enabled_sections()
        # then
        expected = {Character.UpdateSection.ASSETS}
        sections = set(result.values_list("section", flat=True))
        self.assertSetEqual(sections, expected)



"""Tool for clearing celery locks."""

import argparse

from redis import Redis


def main():
    """Clear all celery once locks (if any exist)."""
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    parser.add_argument(
        "--redis-url", default="redis://127.0.0.1:6379/1", help="REDIS URL"
    )
    parser.add_argument("--app-label", default="memberaudit", help="Django app label")
    args = parser.parse_args()

    r = Redis.from_url(args.redis_url)
    if keys := r.keys(f":?:qo_{args.app_label}.*"):
        deleted_count = r.delete(*keys)
    else:
        deleted_count = 0
    print(f"Removed {deleted_count} stuck celery once keys")


if __name__ == "__main__":
    main()



"""Character and CharacterUpdateStatus models."""

# pylint: disable = too-many-lines

import datetime as dt
import hashlib
import json
from dataclasses import dataclass
from typing import Any, Callable, Dict, List, Optional, Set

from bravado.exception import HTTPInternalServerError

from django.contrib.auth.models import User
from django.core.exceptions import ObjectDoesNotExist
from django.core.serializers.json import DjangoJSONEncoder
from django.db import models
from django.utils.functional import cached_property
from django.utils.timezone import now
from django.utils.translation import gettext_lazy as _
from esi.errors import TokenError
from esi.models import Token
from eveuniverse.models import EveEntity

from allianceauth.authentication.models import CharacterOwnership
from allianceauth.eveonline.models import EveCharacter
from allianceauth.notifications import notify
from allianceauth.services.hooks import get_extension_logger
from app_utils.logging import LoggerAddTag

from memberaudit import __title__
from memberaudit.app_settings import (
    MEMBERAUDIT_APP_NAME,
    MEMBERAUDIT_FEATURE_ROLES_ENABLED,
    MEMBERAUDIT_NOTIFY_TOKEN_ERRORS,
    MEMBERAUDIT_SECTION_STALE_MINUTES_CONFIG,
    MEMBERAUDIT_SECTION_STALE_MINUTES_GLOBAL_DEFAULT,
    MEMBERAUDIT_SECTION_STALE_MINUTES_SECTION_DEFAULTS,
)
from memberaudit.constants import EveGroupId
from memberaudit.errors import TokenDoesNotExist
from memberaudit.helpers import UpdateSectionResult
from memberaudit.managers.characters import (
    CharacterManager,
    CharacterUpdateStatusManager,
)
from memberaudit.models._helpers import (
    AddGenericReprMixin,
    store_character_data_to_disk_when_enabled,
)

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


@dataclass(frozen=True)
class _CharacterNeedsUpdate:
    """An object which knows what character sections need to be updated."""

    section_map: Dict["Character.UpdateSection", bool]
    character: "Character"

    def __bool__(self) -> bool:
        """Return True if any section needs to be updated, else False."""
        return any(self.section_map.values())

    def for_section(self, section: "Character.UpdateSection") -> bool:
        """Return True if the given section needs to be updated, else False."""
        return self.section_map[Character.UpdateSection(section)]


class Character(
    AddGenericReprMixin, models.Model
):  # pylint: disable=too-many-public-methods
    """A character in Eve Online managed by Member Audit."""

    # TODO: Maybe move this enum to outside, so it can be imported more easily
    # by other modules
    class UpdateSection(models.TextChoices):
        """A section of content for a character that can be updated separately."""

        ASSETS = "assets", _("assets")
        ATTRIBUTES = "attributes", _("attributes")
        CHARACTER_DETAILS = "character_details", ("character details")
        CONTACTS = "contacts", _("contacts")
        CONTRACTS = "contracts", _("contracts")
        CORPORATION_HISTORY = "corporation_history", _("corporation history")
        FW_STATS = "fw_stats", _("faction warfare statistics")
        IMPLANTS = "implants", _("implants")
        JUMP_CLONES = "jump_clones", _("jump clones")
        LOCATION = "location", _("location")
        LOYALTY = "loyalty", _("loyalty")
        MAILS = "mails", _("mails")
        MINING_LEDGER = "mining_ledger", _("mining ledger")
        ONLINE_STATUS = "online_status", _("online status")
        PLANETS = "planets", _("planets")
        ROLES = "roles", _("roles")
        SHIP = "ship", _("ship")
        SKILLS = "skills", _("skills")
        SKILL_QUEUE = "skill_queue", _("skill queue")
        SKILL_SETS = "skill_sets", _("skill sets")
        STANDINGS = "standings", _("standings")
        TITLES = "titles", _("titles")
        WALLET_BALLANCE = "wallet_balance", _("wallet balance")
        WALLET_JOURNAL = "wallet_journal", _("wallet journal")
        WALLET_TRANSACTIONS = "wallet_transactions", _("wallet transactions")

        @property
        def method_name(self) -> str:
            """Return name of update method corresponding with this section."""
            return f"update_{self.value}"

        @classmethod
        def enabled_sections(cls) -> Set["Character.UpdateSection"]:
            """Return enabled sections."""
            sections = set(Character.UpdateSection)
            if not MEMBERAUDIT_FEATURE_ROLES_ENABLED:
                sections.discard(Character.UpdateSection.ROLES)
            return sections

        @classmethod
        def time_until_section_updates_are_stale(
            cls,
        ) -> Dict["Character.UpdateSection", int]:
            """Return map of each section and their time until an update is stale."""
            config = {
                section: MEMBERAUDIT_SECTION_STALE_MINUTES_GLOBAL_DEFAULT
                for section in cls
            }

            section_defaults = MEMBERAUDIT_SECTION_STALE_MINUTES_SECTION_DEFAULTS
            for key, value in section_defaults.items():
                config[cls(key)] = value

            section_config = MEMBERAUDIT_SECTION_STALE_MINUTES_CONFIG
            for key, value in section_config.items():
                try:
                    config[cls(key)] = value
                except ValueError:
                    logger.warning(
                        "Ignoring invalid section name '%s' in config "
                        "for stale minutes. "
                        "Please correct the invalid config.",
                        value,
                    )

            return config

    class TotalUpdateStatus(models.TextChoices):
        """An summary update status of a character
        representing the update status of all sections.
        """

        DISABLED = "disabled", _("disabled")
        ERROR = "error", _("error")
        LIMITED_TOKEN = "limited_token", _("limited token")
        INCOMPLETE = "incomplete", _("incomplete")
        IN_PROGRESS = "in_progress", _("in progress")
        OK = "ok", _("ok")

        def has_issue(self) -> bool:
            """Return True when status is representing an issue."""
            return self in {
                self.DISABLED,
                self.ERROR,
                self.INCOMPLETE,
                self.LIMITED_TOKEN,
            }

        def bootstrap_style_class(self) -> str:
            """Return bootstrap corresponding bootstrap style class."""
            my_map = {
                self.DISABLED: "text-muted",
                self.ERROR: "text-danger",
                self.IN_PROGRESS: "text-info",
                self.INCOMPLETE: "text-warning",
                self.LIMITED_TOKEN: "text-warning",
                self.OK: "text-success",
            }
            return my_map.get(self, "")

        def description(self) -> str:
            """Return description for an enum object."""
            my_map = {
                self.DISABLED: _("Update is disabled"),
                self.ERROR: _("Errors occurred during update"),
                self.IN_PROGRESS: _("Update is in progress"),
                self.INCOMPLETE: _("One or more sections have not yet been updated"),
                self.LIMITED_TOKEN: _(
                    "One section can not be updated due to a token error"
                ),
                self.OK: _("Update completed successfully"),
            }
            return my_map.get(self, "")

    id = models.AutoField(primary_key=True)
    eve_character = models.OneToOneField(
        EveCharacter,
        related_name="memberaudit_character",
        on_delete=models.CASCADE,
        verbose_name=_("eve character"),
    )

    created_at = models.DateTimeField(
        auto_now_add=True, db_index=True, verbose_name=_("created at")
    )
    is_shared = models.BooleanField(
        default=False,
        verbose_name=_("is shared"),
        help_text="Shared characters can be viewed by recruiters",
    )
    is_disabled = models.BooleanField(
        default=False,
        verbose_name=_("is disabled"),
        help_text="Disabled characters are no longer updated from ESI.",
    )
    mailing_lists = models.ManyToManyField(
        "MailEntity", related_name="characters", verbose_name=_("mailing lists")
    )
    token_error_notified_at = models.DateTimeField(
        default=None,
        null=True,
        help_text=_("Time when user was last notified about a token error."),
    )

    objects = CharacterManager()

    class Meta:
        default_permissions = ()
        verbose_name = _("character")
        verbose_name_plural = _("characters")

    def __str__(self) -> str:
        try:
            return f"{self.eve_character.character_name} (ID:{self.id})"
        except EveCharacter.DoesNotExist:
            return f"Character ID {self.id}"

    # def __repr__(self) -> str:
    #     return f"Character(pk={self.pk}, eve_character='{self.eve_character}')"

    def save(self, *args, **kwargs: dict):
        ignore_cache = kwargs.pop("ignore_cache", False)  # needed for NoSocketsTestCase
        super().save(*args, **kwargs)
        if self.pk and not ignore_cache:  # clear this object from cache if it is there
            self.clear_cache()

    @cached_property
    def name(self) -> str:
        """Return the name of this character."""
        return str(self.eve_character.character_name)

    @cached_property
    def character_ownership(self) -> Optional[CharacterOwnership]:
        """Return the character ownership object of this character."""
        try:
            return self.eve_character.character_ownership
        except ObjectDoesNotExist:
            return None

    @cached_property
    def user(self) -> Optional[User]:
        """Return the user this character belongs to or None."""
        try:
            return self.character_ownership.user
        except AttributeError:
            return None

    @cached_property
    def main_character(self) -> Optional[EveCharacter]:
        """Return the main character related to this character or None."""
        try:
            return self.character_ownership.user.profile.main_character
        except AttributeError:
            return None

    @cached_property
    def is_main(self) -> bool:
        """Return True if this character is a main character, else False."""
        try:
            return self.main_character.character_id == self.eve_character.character_id
        except AttributeError:
            return False

    @cached_property
    def is_orphan(self) -> bool:
        """Return True if this character is an orphan else False.

        An orphan is a character that is not owned anymore by a user."""
        return self.character_ownership is None

    def details_or_none(self):
        """Return character details or None if it does not exist."""
        try:
            return self.details
        except ObjectDoesNotExist:
            return None

    def user_is_owner(self, user: User) -> bool:
        """Return True if the given user is owner of this character"""
        try:
            return self.user == user
        except AttributeError:
            return False

    def user_has_scope(self, user: User) -> bool:
        """Returns True if the given user has the permission to access this character."""
        try:
            if self.user == user:  # shortcut for better performance
                return True
        except AttributeError:
            pass
        return Character.objects.user_has_scope(user).filter(pk=self.pk).exists()

    def user_has_access(self, user: User) -> bool:
        """Returns True if given user has permission to access this character
        in the character viewer.
        """
        try:
            if self.user == user:  # shortcut for better performance
                return True
        except AttributeError:
            pass
        return Character.objects.user_has_access(user).filter(pk=self.pk).exists()

    def has_token_issue(self) -> bool:
        """Return True if character has run into a token error during update, else False."""
        return (
            self.update_status_set.filter_enabled_sections()
            .filter(has_token_error=True)
            .exists()
        )

    def calc_total_update_status(self) -> TotalUpdateStatus:
        """Calculate and return the total update status of this character."""
        if self.is_disabled:
            return self.TotalUpdateStatus.DISABLED

        qs = Character.objects.filter(pk=self.pk).annotate_total_update_status()
        total_update_status = list(qs.values_list("total_update_status", flat=True))[0]
        return self.TotalUpdateStatus(total_update_status)

    def is_update_status_ok(self) -> Optional[bool]:
        """Return summary status of last update for this character.

        Returns:
        - True: if update was complete and without errors
        - False if there where any errors
        - None: if last update is incomplete
        """
        status = self.calc_total_update_status()
        if status == Character.TotalUpdateStatus.OK:
            return True
        if status == Character.TotalUpdateStatus.ERROR:
            return False
        return None

    def reset_token_error_notified_if_status_ok(self):
        """Reset last notification on token error when update is OK again."""
        if self.token_error_notified_at:
            if self.calc_total_update_status() == Character.TotalUpdateStatus.OK:
                self.token_error_notified_at = None
                self.save(update_fields=["token_error_notified_at"])

    def calc_update_needed(self) -> _CharacterNeedsUpdate:
        """Return map of section and if they need to be update."""
        sections_needs_update = {
            section: True for section in self.UpdateSection.enabled_sections()
        }
        status_all: List[CharacterUpdateStatus] = self.update_status_set.all()
        sections_current = {
            obj.section: obj.is_update_needed()
            for obj in status_all
            if obj.section in sections_needs_update
        }
        sections_needs_update.update(sections_current)
        return _CharacterNeedsUpdate(sections_needs_update, self)

    def update_status_for_section(
        self, section: UpdateSection
    ) -> Optional["CharacterUpdateStatus"]:
        "Return update status for a section when it exists or None."
        section = self.UpdateSection(section)
        try:
            return self.update_status_set.get(section=section)
        except CharacterUpdateStatus.DoesNotExist:
            return None

    def is_update_needed_for_section(self, section: UpdateSection) -> bool:
        """Return True if the given section needs to be updated,
        else False.
        """
        status = self.update_status_for_section(section)
        if not status:
            return True
        return status.is_update_needed()

    def has_section_changed(
        self, section: UpdateSection, content: Any, hash_num: int = 1
    ) -> bool:
        """Return False if the content hash for this character's section
        has not changed, else return True.
        """
        status = self.update_status_for_section(section)
        if not status:
            return True
        return status.has_changed(content=content, hash_num=hash_num)

    def update_section_content_hash(
        self, section: UpdateSection, content: Any, hash_num: int = 1
    ) -> None:
        """Update hash for a section."""
        section_obj: CharacterUpdateStatus = self.update_status_set.get_or_create(
            character=self, section=section
        )[0]
        section_obj.update_content_hash(content=content, hash_num=hash_num)

    def reset_update_section(self, section: UpdateSection) -> "CharacterUpdateStatus":
        """Reset status of given update section and returns it."""
        update_status_obj: CharacterUpdateStatus = self.update_status_set.get_or_create(
            section=section
        )[0]
        update_status_obj.reset()
        return update_status_obj

    def update_status_as_dict(self) -> Dict[str, Any]:
        """Return current update status for this character as dict."""
        return {obj.section: obj for obj in self.update_status_set.all()}

    def update_section_if_changed(
        self,
        section: UpdateSection,
        fetch_func: Callable,
        store_func: Optional[Callable],
        force_update: bool = False,
        hash_num: int = 1,
    ) -> UpdateSectionResult:
        """Update a character's section from ESI if it has changed or is forced.

        Args:
            - section: Section this update is related to
            - fetch_func: A function that fetched the data from ESI
            - store_func: A function that stored the data in the DB.
                This can be skipped by providing None.
                store_func can optionally return a list of entity IDs to resolve
            - forced_update: Data will always be stored when set to True
            - hash_num: To access sub-sections by ID

        Returns:
            - A tuple of the result data or None if data is unchanged
                and a flag that is True if data was changed,
                False when it was not change, else None
        """
        section = self.UpdateSection(section)
        try:
            data = fetch_func(character=self)
        except HTTPInternalServerError as ex:
            # handle the occasional occurring http 500 error from this endpoint
            logger.warning(
                "%s: Received an HTTP internal server error "
                "when trying to fetch %s: %s ",
                self,
                section,
                ex,
            )
            return UpdateSectionResult(is_changed=None, is_updated=False)

        store_character_data_to_disk_when_enabled(
            character=self,
            data=data,
            section=section,
            suffix=hash_num if hash_num > 1 else "",
        )

        is_changed = self.has_section_changed(
            section=section, content=data, hash_num=hash_num
        )
        if not force_update and not is_changed:
            logger.info("%s: %s has not changed", self, section.label)
            return UpdateSectionResult(is_changed=is_changed, is_updated=False)

        if store_func:
            ids_to_resolve = store_func(self, data)
            is_updated = True
            if ids_to_resolve:
                logger.debug(
                    "%s: Received %d EveEntity IDs from %s "
                    "which might need to be resolved",
                    self,
                    len(ids_to_resolve),
                    section,
                )
                EveEntity.objects.bulk_resolve_ids(ids_to_resolve)
        else:
            is_updated = False

        self.update_section_content_hash(
            section=section, content=data, hash_num=hash_num
        )
        return UpdateSectionResult(
            is_changed=is_changed, is_updated=is_updated, data=data
        )

    def update_section_log_result(
        self,
        section: UpdateSection,
        is_success: bool,
        is_updated: bool = False,
        error_message: str = None,
    ) -> None:
        """Log update result for a character's section."""
        error_message = error_message if error_message else ""
        defaults = {
            "is_success": is_success,
            "has_token_error": False,
            "error_message": error_message,
            "run_finished_at": now(),
        }
        obj: CharacterUpdateStatus = self.update_status_set.update_or_create(
            section=section, defaults=defaults
        )[0]
        if is_updated:
            obj.update_started_at = obj.run_started_at
            obj.update_finished_at = now()
            obj.save()

        status = "successfully" if is_success else "with errors"
        logger.info("%s: %s update run completed %s", self, section.label, status)

    def perform_update_with_error_logging(
        self, section: UpdateSection, method: Callable, *args, **kwargs
    ) -> UpdateSectionResult:
        """Facilitate catching and logging of exceptions potentially occurring
        during a character update.
        """
        try:
            result = method(*args, **kwargs)
        except Exception as ex:
            error_message = f"{type(ex).__name__}: {str(ex)}"
            is_token_error = isinstance(ex, (TokenError))
            logger.error(
                "%s: %s: Error ocurred: %s",
                self,
                section.label,
                error_message,
                exc_info=not is_token_error,  # hide details when token error
            )
            self.update_status_set.update_or_create(
                character=self,
                section=section,
                defaults={
                    "is_success": False,
                    "has_token_error": is_token_error,
                    "error_message": error_message,
                    "run_finished_at": now(),
                },
            )
            raise ex

        if not isinstance(result, UpdateSectionResult):
            raise TypeError(
                f"{method}: Method must have return type: UpdateSectionResult"
            )

        return result

    def fetch_token(self, scopes=None) -> Token:
        """Return a valid token for this character and scope.

        Args:
            - scopes: Optionally provide the required scopes.
            Otherwise will use all scopes defined for this character.

        Exceptions:
            - TokenDoesNotExist: If no matching token is found
            - TokenError: Various token errors
        """
        if self.is_orphan:
            raise TokenError(f"Orphaned characters have no token: {self}") from None

        token = (
            Token.objects.prefetch_related("scopes")
            .filter(user=self.user, character_id=self.eve_character.character_id)
            .require_scopes(scopes if scopes else self.get_esi_scopes())
            .require_valid()
            .first()
        )
        if not token:
            if (
                MEMBERAUDIT_NOTIFY_TOKEN_ERRORS
                and self.user
                and not self.token_error_notified_at
            ):
                title = (
                    f"{__title__}: Invalid or missing token for {self.eve_character}"
                )
                message = (
                    f"{MEMBERAUDIT_APP_NAME} could not find a valid token for your "
                    f"character {self.eve_character}.\n"
                    f"Please re-add that character to {MEMBERAUDIT_APP_NAME} "
                    "at your earliest convenience to update your token."
                )
                notify.danger(user=self.user, title=title, message=message)
                self.token_error_notified_at = now()
                self.save(update_fields=["token_error_notified_at"])

            raise TokenDoesNotExist(
                f"Could not find a matching token for {self} with scopes: {scopes}."
            )
        return token

    def assets_build_list_from_esi(self, force_update=False) -> UpdateSectionResult:
        """Fetch character's assets from ESI and return it.

        Or return None if no update is required.
        """
        return self.assets.fetch_from_esi(self, force_update)

    def assets_preload_objects(self, asset_list: list) -> UpdateSectionResult:
        """Preload objects needed to build the character's asset tree from ESI."""
        return self.assets.preload_objects_from_esi(self, asset_list)

    def update_attributes(self, force_update: bool = False) -> UpdateSectionResult:
        """Update the character's learning attributes from ESI."""
        from memberaudit.models import CharacterAttributes

        return CharacterAttributes.objects.update_or_create_esi(self, force_update)

    def update_character_details(
        self, force_update: bool = False
    ) -> UpdateSectionResult:
        """Update the character's details from ESI."""
        from memberaudit.models import CharacterDetails

        return CharacterDetails.objects.update_or_create_esi(self, force_update)

    def update_contact_labels(self, force_update: bool = False) -> UpdateSectionResult:
        """Update character's contact labels from ESI."""
        return self.contact_labels.update_or_create_esi(self, force_update)

    def update_contacts(self, force_update: bool = False) -> UpdateSectionResult:
        """Update character's contacts from ESI."""
        return self.contacts.update_or_create_esi(self, force_update)

    def update_contract_headers(
        self, force_update: bool = False
    ) -> UpdateSectionResult:
        """Update the character's contract headers from ESI."""
        return self.contracts.update_or_create_esi(self, force_update)

    def update_contract_items(self, contract) -> UpdateSectionResult:
        """Update the character's contract items from ESI."""
        return contract.items.update_or_create_esi(self, contract)

    def update_contract_bids(self, contract) -> UpdateSectionResult:
        """Update the character's contract bids from ESI."""
        return contract.bids.update_or_create_esi(self, contract)

    def update_corporation_history(
        self, force_update: bool = False
    ) -> UpdateSectionResult:
        """Update the character's corporation history from ESI."""
        return self.corporation_history.update_or_create_esi(self, force_update)

    def update_fw_stats(self, force_update: bool = False) -> UpdateSectionResult:
        """Update character's FW stats from ESI."""
        from memberaudit.models import CharacterFwStats

        return CharacterFwStats.objects.update_or_create_esi(self, force_update)

    def update_implants(self, force_update: bool = False) -> UpdateSectionResult:
        """Update the character's current implants from ESI."""
        return self.implants.update_or_create_esi(self, force_update)

    def update_location(self, force_update: bool = False) -> UpdateSectionResult:
        """Update the character's current location from ESI."""
        from memberaudit.models import CharacterLocation

        return CharacterLocation.objects.update_or_create_esi(self, force_update)

    def update_loyalty(self, force_update: bool = False) -> UpdateSectionResult:
        """Update the character's loyalty entries from ESI."""
        return self.loyalty_entries.update_or_create_esi(self, force_update)

    def update_jump_clones(self, force_update: bool = False) -> UpdateSectionResult:
        """Update the character's jump clones from ESI."""
        return self.jump_clones.update_or_create_esi(self, force_update)

    def update_mailing_lists(self, force_update: bool = False) -> UpdateSectionResult:
        """Update the character's mailing lists from ESI."""
        return self.mailing_lists.update_or_create_mailing_lists_esi(self, force_update)

    def update_mail_labels(self, force_update: bool = False) -> UpdateSectionResult:
        """Update the character's mail labels from ESI."""
        return self.mail_labels.update_or_create_esi(self, force_update)

    def update_mail_headers(self, force_update: bool = False) -> UpdateSectionResult:
        """Update the character's mail headers from ESI."""
        return self.mails.update_or_create_headers_esi(self, force_update)

    def update_mail_body(self, mail, force_update: bool = False) -> UpdateSectionResult:
        """Update the characters's mail body from ESI."""
        return self.mails.update_or_create_body_esi(self, mail, force_update)

    def update_mining_ledger(self, force_update: bool = False) -> UpdateSectionResult:
        """Update character's mining ledger from ESI."""
        return self.mining_ledger.update_or_create_esi(self, force_update)

    def update_online_status(self, force_update: bool = False) -> UpdateSectionResult:
        """Update the character's online status from ESI."""
        from memberaudit.models import CharacterOnlineStatus

        return CharacterOnlineStatus.objects.update_or_create_esi(self, force_update)

    def update_planets(self, force_update: bool = False) -> UpdateSectionResult:
        """Update the character's planets from ESI."""
        return self.planets.update_or_create_esi(self, force_update)

    def update_roles(self, force_update: bool = False) -> UpdateSectionResult:
        """Update the character's corporation roles from ESI."""
        return self.roles.update_or_create_esi(self, force_update)

    def update_titles(self, force_update: bool = False) -> UpdateSectionResult:
        """Update the character's corporation titles from ESI."""
        return self.titles.update_or_create_esi(self, force_update)

    def update_ship(self, force_update: bool = False) -> UpdateSectionResult:
        """Update the character's current ship from ESI."""
        from memberaudit.models import CharacterShip

        return CharacterShip.objects.update_or_create_esi(self, force_update)

    def update_skill_queue(self, force_update: bool = False) -> UpdateSectionResult:
        """Update the character's skill queue from ESI."""
        return self.skillqueue.update_or_create_esi(self, force_update)

    def update_skill_sets(self) -> UpdateSectionResult:
        """Update the character's skill sets."""
        return self.skill_set_checks.update_for_character(self)

    def update_skills(self, force_update: bool = False) -> UpdateSectionResult:
        """Update the character's skills from ESI."""
        return self.skills.update_or_create_esi(self, force_update)

    def update_standings(self, force_update: bool = False) -> UpdateSectionResult:
        """Update the character's standings from ESI."""
        return self.standings.update_or_create_esi(self, force_update)

    def update_wallet_balance(self, force_update: bool = False) -> UpdateSectionResult:
        """Update the character's wallet balance from ESI."""
        from memberaudit.models import CharacterWalletBalance

        return CharacterWalletBalance.objects.update_or_create_esi(self, force_update)

    def update_wallet_journal(self, force_update: bool = False) -> UpdateSectionResult:
        """Update the character's wallet journal from ESI."""
        return self.wallet_journal.update_or_create_esi(self, force_update)

    def update_wallet_transactions(
        self, force_update: bool = False
    ) -> UpdateSectionResult:
        """Update the character's wallet transactions from ESI."""
        return self.wallet_transactions.update_or_create_esi(self, force_update)

    def update_sharing_consistency(self):
        """Update sharing to ensure consistency with permissions."""
        if (
            self.is_shared
            and self.user
            and not self.user.has_perm("memberaudit.share_characters")
        ):
            self.is_shared = False
            self.save()
            logger.info(
                "%s: Unshared this character, "
                "because it's owner no longer has the permission to share characters.",
                self,
            )

    def generate_asset_from_current_ship_and_location(self) -> Optional[dict]:
        """Return generated asset item record from current ship and location
        or None it can not be generated.
        """
        from .character_sections_2 import CharacterLocation
        from .character_sections_3 import CharacterShip
        from .general import Location

        try:
            ship: CharacterShip = CharacterShip.objects.select_related("eve_type").get(
                character_id=self.id
            )
        except CharacterShip.DoesNotExist:
            return None

        if not ship.item_id:
            return None  # item ID is 0 from the migration

        if ship.eve_type.eve_group_id == EveGroupId.CAPSULE:
            return None  # we don't add capsules

        try:
            character_location: CharacterLocation = (
                CharacterLocation.objects.select_related(
                    "eve_solar_system",
                    "location",
                    "location__eve_solar_system",
                    "location__eve_type",
                ).get(character_id=self.id)
            )
            location = character_location.location_safe()
        except CharacterLocation.DoesNotExist:
            location, _ = Location.objects.get_or_create_unknown_location()

        ship_asset_record = {
            "is_blueprint_copy": False,
            "is_singleton": True,
            "item_id": ship.item_id,
            "location_flag": "Hangar",
            "location_id": location.id,
            "location_type": location.asset_location_type(),
            "name": ship.name,
            "quantity": 1,
            "type_id": ship.eve_type.id,
        }
        return ship_asset_record

    def clear_cache(self) -> None:
        """Remove this character from cache."""
        Character.objects.clear_cache(pk=self.pk)

    @staticmethod
    def get_esi_scopes() -> List[str]:
        """Return all enabled ESI scopes required to update this character."""
        scopes = [
            "esi-assets.read_assets.v1",
            "esi-bookmarks.read_character_bookmarks.v1",
            "esi-calendar.read_calendar_events.v1",
            "esi-characters.read_agents_research.v1",
            "esi-characters.read_blueprints.v1",
            "esi-characters.read_contacts.v1",
            "esi-characters.read_corporation_roles.v1",
            "esi-characters.read_fatigue.v1",
            "esi-characters.read_fw_stats.v1",
            "esi-characters.read_loyalty.v1",
            "esi-characters.read_medals.v1",
            "esi-characters.read_notifications.v1",
            "esi-characters.read_opportunities.v1",
            "esi-characters.read_standings.v1",
            "esi-characters.read_titles.v1",
            "esi-clones.read_clones.v1",
            "esi-clones.read_implants.v1",
            "esi-contracts.read_character_contracts.v1",
            "esi-corporations.read_corporation_membership.v1",
            "esi-industry.read_character_jobs.v1",
            "esi-industry.read_character_mining.v1",
            "esi-killmails.read_killmails.v1",
            "esi-location.read_location.v1",
            "esi-location.read_online.v1",
            "esi-location.read_ship_type.v1",
            "esi-mail.organize_mail.v1",
            "esi-mail.read_mail.v1",
            "esi-markets.read_character_orders.v1",
            "esi-markets.structure_markets.v1",
            "esi-planets.manage_planets.v1",
            "esi-planets.read_customs_offices.v1",
            "esi-search.search_structures.v1",
            "esi-skills.read_skillqueue.v1",
            "esi-skills.read_skills.v1",
            "esi-universe.read_structures.v1",
            "esi-wallet.read_character_wallet.v1",
        ]
        return sorted(scopes)


class CharacterUpdateStatus(AddGenericReprMixin, models.Model):
    """An object for tracking the update status of a character's section."""

    character = models.ForeignKey(
        Character, on_delete=models.CASCADE, related_name="update_status_set"
    )
    section = models.CharField(
        max_length=64, choices=Character.UpdateSection.choices, db_index=True
    )

    content_hash_1 = models.CharField(max_length=32, default="")
    content_hash_2 = models.CharField(max_length=32, default="")
    content_hash_3 = models.CharField(max_length=32, default="")
    has_token_error = models.BooleanField(
        default=False, help_text="Whether this section has a token error."
    )
    is_success = models.BooleanField(
        null=True,
        default=None,
        db_index=True,
    )
    error_message = models.TextField()
    run_finished_at = models.DateTimeField(
        null=True,
        default=None,
        db_index=True,
        help_text="Finish time of an update attempt",
    )
    run_started_at = models.DateTimeField(
        null=True,
        default=None,
        db_index=True,
        help_text="Start time of an update attempt",
    )
    update_finished_at = models.DateTimeField(
        null=True,
        default=None,
        db_index=True,
        help_text="Finish time of an actual update",
    )
    update_started_at = models.DateTimeField(
        null=True,
        default=None,
        db_index=True,
        help_text="Start time of an actual update",
    )

    objects = CharacterUpdateStatusManager()

    class Meta:
        default_permissions = ()
        constraints = [
            models.UniqueConstraint(
                fields=["character", "section"],
                name="functional_pk_charactersyncstatus",
            )
        ]
        verbose_name = _("character update status")
        verbose_name_plural = _("character update status")

    def __str__(self) -> str:
        return f"{self.character}-{self.section}"

    @property
    def is_enabled(self) -> bool:
        """Return True if this section is currently enabled."""
        return self.section in Character.UpdateSection.enabled_sections()

    @property
    def is_updating(self) -> bool:
        """Return True if this section is currently being updated."""
        if not self.run_started_at and not self.run_finished_at:
            return False

        return self.run_started_at is not None and self.run_finished_at is None

    def has_changed(self, content: Any, hash_num: int = 1) -> bool:
        """Return True when given content is not the same as previous one, else False.

        Specify optionally which sub section to update via the hash_num (1, 2 or 3).
        """
        new_hash = self._calculate_hash(content)
        if hash_num == 2:
            content_hash = self.content_hash_2
        elif hash_num == 3:
            content_hash = self.content_hash_3
        else:
            content_hash = self.content_hash_1

        result = new_hash != content_hash
        return result

    def update_content_hash(self, content: Any, hash_num: int = 1):
        """Update content hash for this update status.

        Specify optionally which sub section to update via the hash_num (1, 2 or 3).
        """
        new_hash = self._calculate_hash(content)
        if hash_num == 2:
            self.content_hash_2 = new_hash
        elif hash_num == 3:
            self.content_hash_3 = new_hash
        else:
            self.content_hash_1 = new_hash

        self.save()

    def is_update_needed(self) -> bool:
        """Return True if this status is stale and needs to be updated, else False.
        But never report sections with token error as stale.
        """
        section = Character.UpdateSection(self.section)
        if not self.is_success or not self.run_finished_at:
            needs_update = True
        else:
            minutes = section_time_until_stale[section]
            deadline = now() - dt.timedelta(minutes=minutes)
            needs_update = self.run_finished_at <= deadline

        if needs_update and self.has_token_error:
            logger.warning(
                "%s: Ignoring update need because of token error for section: %s",
                self.character,
                section.label,
            )
            return False

        return needs_update

    @staticmethod
    def _calculate_hash(content: Any) -> str:
        return hashlib.md5(
            json.dumps(content, cls=DjangoJSONEncoder).encode("utf-8")
        ).hexdigest()

    def reset(self) -> None:
        """Reset this update status."""
        self.is_success = None
        self.error_message = ""
        self.has_token_error = False
        self.run_started_at = now()
        self.run_finished_at = None
        self.save()
        # TODO: Check if the hash also needs to be reset?


section_time_until_stale = (
    Character.UpdateSection.time_until_section_updates_are_stale()
)
"""Mapping of all sections with effective stale minutes."""


def enabled_sections_by_stale_minutes() -> List[Character.UpdateSection]:
    """Return list of enabled sections, ordered by stale minutes ascending."""
    sections_and_minutes = {
        section: value
        for section, value in section_time_until_stale.items()
        if section in Character.UpdateSection.enabled_sections()
    }
    sections = [
        obj[0] for obj in sorted(sections_and_minutes.items(), key=lambda o: o[1])
    ]

    return sections



from unittest.mock import patch

from django.test import TestCase, override_settings

from app_utils.esi import EsiStatus

from memberaudit import tasks
from memberaudit.tests.testdata.esi_client_stub import esi_stub
from memberaudit.tests.testdata.load_entities import load_entities
from memberaudit.tests.testdata.load_eveuniverse import load_eveuniverse
from memberaudit.tests.testdata.load_locations import load_locations
from memberaudit.tests.utils import (
    create_memberaudit_character,
    reset_celery_once_locks,
)

MANAGERS_PATH = "memberaudit.managers"
MODELS_PATH = "memberaudit.models"
TASKS_PATH = "memberaudit.tasks"


@patch(MANAGERS_PATH + ".general.fetch_esi_status", lambda: EsiStatus(True, 99, 60))
@patch(MANAGERS_PATH + ".character_sections_1.data_retention_cutoff", lambda: None)
@patch(MANAGERS_PATH + ".character_sections_2.data_retention_cutoff", lambda: None)
@patch(MANAGERS_PATH + ".character_sections_3.data_retention_cutoff", lambda: None)
@patch(MANAGERS_PATH + ".character_sections_1.esi", esi_stub)
@patch(MANAGERS_PATH + ".character_sections_2.esi", esi_stub)
@patch(MANAGERS_PATH + ".character_sections_3.esi", esi_stub)
@patch(MANAGERS_PATH + ".general.esi", esi_stub)
@override_settings(
    CELERY_ALWAYS_EAGER=True,
    CELERY_EAGER_PROPAGATES_EXCEPTIONS=True,
    APP_UTILS_OBJECT_CACHE_DISABLED=True,
)
class TestTasksIntegration(TestCase):
    fixtures = ["disable_analytics.json"]

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()
        load_entities()
        load_locations()
        reset_celery_once_locks()

    def test_should_update_all_characters(self):
        # given
        character_1001 = create_memberaudit_character(1001)
        # when
        tasks.update_all_characters()
        # then
        self.assertTrue(character_1001.is_update_status_ok())



"""Logic for converting Eve Online XML into other formats."""

import unicodedata

import bleach
import bs4

from eveuniverse.core import evexml

from allianceauth.services.hooks import get_extension_logger
from app_utils.logging import LoggerAddTag

from memberaudit import __title__

logger = LoggerAddTag(get_extension_logger(__name__), __title__)

DEFAULT_FONT_SIZE = 13

EVEXML_ALLOWED_TAGS = [
    "a",
    "b",
    "br",
    "font",
    "i",
    "u",
]

EVEXML_ALLOWED_ATTRIBUTES = {
    "a": ["href"],
    "font": ["size"],
}

EVEXML_ALLOWED_PROTOCOLS = {
    "http",
    "https",
    "showinfo",
    "killreport",
}


def eve_xml_to_html(xml_doc: str, add_default_style: bool = False) -> str:
    """Convert Eve Online XML to HTML.

    Args:
    - xml_doc: XML document
    - add_default_style: When set true will add the default style to all unstyled fragments
    """
    xml_doc = unicodedata.normalize("NFKC", xml_doc)
    xml_doc = bleach.clean(
        xml_doc,
        protocols=EVEXML_ALLOWED_PROTOCOLS,
        tags=EVEXML_ALLOWED_TAGS,
        attributes=EVEXML_ALLOWED_ATTRIBUTES,
        strip=True,
    )
    soup = bs4.BeautifulSoup(xml_doc, "html.parser")
    _convert_font_tag(soup)
    _convert_a_tag(soup)
    if add_default_style:
        _add_default_style(soup)
    return str(soup)


def _convert_font_tag(soup):
    """Convert the font tags into HTML style."""
    for element in soup.find_all("font"):
        if not element.contents:
            # Element is empty, just remove it from tree as it won't render visibly.
            element.decompose()
            continue
        element.name = "span"
        styles = []
        if "size" in element.attrs:
            # Cast to int and use the absolute value, otherwise use default font size.
            # Prevents a limited CSS injection via this attribute.
            try:
                size = abs(int(element["size"]))
            except ValueError:
                size = DEFAULT_FONT_SIZE
            styles.append(f"font-size: {size}px")
            del element["size"]
        if styles:
            element["style"] = "; ".join(styles)


def _convert_a_tag(soup: bs4.BeautifulSoup):
    """Convert links into HTML."""
    for element in soup.find_all("a"):
        if not element.contents:
            # Element is empty, just remove it from tree as it won't render visibly.
            element.decompose()
            continue
        # Verify if the href attr exists as it can be removed by Bleach for unsupported protocols.
        # Otherwise, we replace this element with its children.
        if not element.get("href"):
            element.replace_with_children()
            continue
        new_href = evexml.eve_link_to_url(element["href"])
        if new_href:
            element["href"] = new_href
            element["target"] = "_blank"
        else:
            element["href"] = "#"


def _add_default_style(soup: bs4.BeautifulSoup):
    """Add default style to all unstyled fragments."""
    for element in soup.children:
        if isinstance(element, bs4.NavigableString):
            new_tag = soup.new_tag("span")
            new_tag["style"] = f"font-size: {DEFAULT_FONT_SIZE}px"
            element.wrap(new_tag)



import datetime as dt
from unittest.mock import MagicMock, patch

from bravado.exception import HTTPForbidden, HTTPNotFound, HTTPUnauthorized
from celery_once import AlreadyQueued

from django.test import TestCase, override_settings
from django.utils.timezone import now
from esi.models import Token
from eveuniverse.models import EveEntity, EveSolarSystem, EveType

from allianceauth.eveonline.models import EveCorporationInfo
from allianceauth.notifications.models import Notification
from app_utils.esi import EsiStatus, fetch_esi_status
from app_utils.esi_testing import BravadoOperationStub, BravadoResponseStub
from app_utils.testing import (
    NoSocketsTestCase,
    create_authgroup,
    create_state,
    create_user_from_evecharacter,
)

from memberaudit.models import (
    ComplianceGroupDesignation,
    Location,
    MailEntity,
    SkillSet,
)
from memberaudit.tests.testdata.constants import EveTypeId
from memberaudit.tests.testdata.esi_client_stub import esi_client_stub
from memberaudit.tests.testdata.factories import (
    create_compliance_group,
    create_fitting,
    create_location,
    create_mail_entity,
    create_skill,
    create_skill_plan,
    create_skill_set_group,
)
from memberaudit.tests.testdata.load_entities import load_entities
from memberaudit.tests.testdata.load_eveuniverse import load_eveuniverse
from memberaudit.tests.utils import (
    add_auth_character_to_user,
    add_memberaudit_character_to_user,
    create_memberaudit_character,
)

MANAGERS_PATH = "memberaudit.managers.general"
TASKS_PATH = "memberaudit.tasks"


@patch(
    "allianceauth.authentication.models.notify", lambda *args, **kwargs: None
)  # state changes trigger notify
@patch(MANAGERS_PATH + ".notify", spec=True)
class TestComplianceGroupDesignation(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_entities()

    def test_should_add_group_to_compliant_user_and_notify(self, mock_notify):
        # given
        compliance_group = create_compliance_group()
        other_group = create_authgroup(internal=True)
        user, _ = create_user_from_evecharacter(
            1001, permissions=["memberaudit.basic_access"]
        )
        add_memberaudit_character_to_user(user, 1001)
        # when
        ComplianceGroupDesignation.objects.update_user(user)
        # then
        self.assertIn(compliance_group, user.groups.all())
        self.assertNotIn(other_group, user.groups.all())
        self.assertEqual(mock_notify.call_count, 1)
        args, kwargs = mock_notify.call_args
        self.assertEqual(kwargs["level"], Notification.Level.SUCCESS)
        self.assertEqual(args[0], user)

    def test_should_add_state_group_to_compliant_user_when_state_matches(
        self, mock_notify
    ):
        # given
        member_corporation = EveCorporationInfo.objects.get(corporation_id=2001)
        my_state = create_state(member_corporations=[member_corporation], priority=200)
        compliance_group = create_compliance_group(states=[my_state])

        user, _ = create_user_from_evecharacter(
            1001, permissions=["memberaudit.basic_access"]
        )
        add_memberaudit_character_to_user(user, 1001)
        # when
        ComplianceGroupDesignation.objects.update_user(user)
        # then
        self.assertIn(compliance_group, user.groups.all())

    def test_should_not_add_state_group_to_compliant_user_when_state_not_matches(
        self, mock_notify
    ):
        # given
        my_state = create_state(priority=200)
        compliance_group = create_compliance_group(states=[my_state])
        user, _ = create_user_from_evecharacter(
            1001, permissions=["memberaudit.basic_access"]
        )
        add_memberaudit_character_to_user(user, 1001)
        # when
        ComplianceGroupDesignation.objects.update_user(user)
        # then
        self.assertNotIn(compliance_group, user.groups.all())
        self.assertFalse(user.notification_set.exists())

    # def test_should_not_notify_if_compliant_but_no_groups_added(self):
    #     # given
    #     member_corporation = EveCorporationInfo.objects.get(corporation_id=2001)
    #     my_state = create_state(member_corporations=[member_corporation], priority=200)
    #     compliance_group = create_compliance_group(states=[my_state])
    #     user, _ = create_user_from_evecharacter(
    #         1001, permissions=["memberaudit.basic_access"]
    #     )
    #     add_memberaudit_character_to_user(user, 1001)
    #     # when
    #     ComplianceGroupDesignation.objects.update_user(user)
    #     # then
    #     self.assertIn(compliance_group, user.groups.all())

    def test_should_add_multiple_groups_to_compliant_user(self, mock_notify):
        # given
        compliance_group_1 = create_compliance_group()
        compliance_group_2 = create_compliance_group()
        user, _ = create_user_from_evecharacter(
            1001, permissions=["memberaudit.basic_access"]
        )
        add_memberaudit_character_to_user(user, 1001)
        # when
        ComplianceGroupDesignation.objects.update_user(user)
        # then
        self.assertIn(compliance_group_1, user.groups.all())
        self.assertIn(compliance_group_2, user.groups.all())

    def test_should_remove_group_from_non_compliant_user_and_notify(self, mock_notify):
        # given
        compliance_group = create_compliance_group()
        other_group = create_authgroup(internal=True)
        user, _ = create_user_from_evecharacter(
            1001, permissions=["memberaudit.basic_access"]
        )
        user.groups.add(compliance_group, other_group)
        # when
        ComplianceGroupDesignation.objects.update_user(user)
        # then
        self.assertNotIn(compliance_group, user.groups.all())
        self.assertIn(other_group, user.groups.all())
        args, kwargs = mock_notify.call_args
        self.assertEqual(kwargs["level"], Notification.Level.WARNING)
        self.assertEqual(args[0], user)

    def test_should_remove_multiple_groups_from_non_compliant_user(self, mock_notify):
        # given
        compliance_group_1 = create_compliance_group()
        compliance_group_2 = create_compliance_group()
        other_group = create_authgroup(internal=True)
        user, _ = create_user_from_evecharacter(
            1001, permissions=["memberaudit.basic_access"]
        )
        user.groups.add(compliance_group_1, compliance_group_2, other_group)
        # when
        ComplianceGroupDesignation.objects.update_user(user)
        # then
        self.assertNotIn(compliance_group_1, user.groups.all())
        self.assertNotIn(compliance_group_2, user.groups.all())
        self.assertIn(other_group, user.groups.all())

    def test_user_with_one_registered_and_one_unregistered_character_is_not_compliant(
        self, mock_notify
    ):
        # given
        compliance_group = create_compliance_group()
        user, _ = create_user_from_evecharacter(
            1001, permissions=["memberaudit.basic_access"]
        )
        add_memberaudit_character_to_user(user, 1001)
        add_auth_character_to_user(user, 1002)
        user.groups.add(compliance_group)
        # when
        ComplianceGroupDesignation.objects.update_user(user)
        # then
        self.assertNotIn(compliance_group, user.groups.all())

    def test_user_without_basic_permission_is_not_compliant(self, mock_notify):
        # given
        compliance_group = create_compliance_group()
        user, _ = create_user_from_evecharacter(1001)
        add_memberaudit_character_to_user(user, 1001)
        user.groups.add(compliance_group)
        # when
        ComplianceGroupDesignation.objects.update_user(user)
        # then
        self.assertNotIn(compliance_group, user.groups.all())

    def test_should_add_missing_groups_if_user_remains_compliant(self, mock_notify):
        # given
        compliance_group_1 = create_compliance_group()
        compliance_group_2 = create_compliance_group()
        other_group = create_authgroup(internal=True)
        user, _ = create_user_from_evecharacter(
            1001, permissions=["memberaudit.basic_access"]
        )
        add_memberaudit_character_to_user(user, 1001)
        user.groups.add(compliance_group_1)
        # when
        ComplianceGroupDesignation.objects.update_user(user)
        # then
        self.assertIn(compliance_group_1, user.groups.all())
        self.assertIn(compliance_group_2, user.groups.all())
        self.assertNotIn(other_group, user.groups.all())
        self.assertEqual(user.notification_set.count(), 0)


class TestMailEntityManager(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_entities()

    def test_get_or_create_esi_1(self):
        """When entity already exists, return it"""
        # given
        create_mail_entity(
            id=1234, category=MailEntity.Category.CHARACTER, name="John Doe"
        )
        # when
        obj, created = MailEntity.objects.get_or_create_esi(id=1234)
        # then
        self.assertFalse(created)
        self.assertEqual(obj.category, MailEntity.Category.CHARACTER)
        self.assertEqual(obj.name, "John Doe")

    @patch(MANAGERS_PATH + ".fetch_esi_status", spec=True)
    def test_get_or_create_esi_2(self, mock_fetch_esi_status):
        """When entity does not exist, create it from ESI / existing EveEntity"""
        mock_fetch_esi_status.return_value = EsiStatus(True, 99, 60)

        obj, created = MailEntity.objects.get_or_create_esi(id=1001)

        self.assertTrue(created)
        self.assertEqual(obj.category, MailEntity.Category.CHARACTER)
        self.assertEqual(obj.name, "Bruce Wayne")

    @patch(MANAGERS_PATH + ".fetch_esi_status", spec=True)
    def test_update_or_create_esi_1(self, mock_fetch_esi_status):
        """When entity does not exist, create it from ESI / existing EveEntity"""
        mock_fetch_esi_status.return_value = EsiStatus(True, 99, 60)

        obj, created = MailEntity.objects.update_or_create_esi(id=1001)

        self.assertTrue(created)
        self.assertEqual(obj.category, MailEntity.Category.CHARACTER)
        self.assertEqual(obj.name, "Bruce Wayne")

    def test_update_or_create_esi_2(self):
        """When entity already exist and is not a mailing list,
        then update it from ESI / existing EveEntity
        """
        create_mail_entity(
            id=1001, category=MailEntity.Category.CHARACTER, name="John Doe"
        )
        obj, created = MailEntity.objects.update_or_create_esi(id=1001)

        self.assertFalse(created)
        self.assertEqual(obj.category, MailEntity.Category.CHARACTER)
        self.assertEqual(obj.name, "Bruce Wayne")

    def test_update_or_create_esi_3(self):
        """When entity already exist and is a mailing list, then do nothing"""
        create_mail_entity(
            id=9001, category=MailEntity.Category.MAILING_LIST, name="Dummy"
        )
        obj, created = MailEntity.objects.update_or_create_esi(id=9001)

        self.assertFalse(created)
        self.assertEqual(obj.category, MailEntity.Category.MAILING_LIST)
        self.assertEqual(obj.name, "Dummy")
        # method must not create an EveEntity object for the mailing list
        self.assertFalse(EveEntity.objects.filter(id=9001).exists())

    @patch(MANAGERS_PATH + ".fetch_esi_status", MagicMock(return_value=MagicMock()))
    def test_update_or_create_esi_4(self):
        """When entity does not exist and is a mailing list, then create it."""
        # when
        with patch(
            MANAGERS_PATH + ".EveEntity.objects.get_or_create_esi", spec=True
        ) as m:
            m.return_value = None, False
            obj, created = MailEntity.objects.update_or_create_esi(id=9001)
        # when
        self.assertTrue(created)
        self.assertEqual(obj.id, 9001)
        self.assertEqual(obj.category, MailEntity.Category.MAILING_LIST)

    def test_update_or_create_from_eve_entity_1(self):
        """When entity does not exist, create it from given EveEntity"""
        eve_entity = EveEntity.objects.get(id=1001)
        obj, created = MailEntity.objects.update_or_create_from_eve_entity(eve_entity)

        self.assertTrue(created)
        self.assertEqual(obj.category, MailEntity.Category.CHARACTER)
        self.assertEqual(obj.name, "Bruce Wayne")

    def test_update_or_create_from_eve_entity_2(self):
        """When entity already exist, update it from given EveEntity"""
        create_mail_entity(
            id=1001, category=MailEntity.Category.CHARACTER, name="John Doe"
        )

        eve_entity = EveEntity.objects.get(id=1001)
        obj, created = MailEntity.objects.update_or_create_from_eve_entity(eve_entity)

        self.assertFalse(created)
        self.assertEqual(obj.category, MailEntity.Category.CHARACTER)
        self.assertEqual(obj.name, "Bruce Wayne")

    def test_update_or_create_from_eve_entity_id_1(self):
        """When entity does not exist, create it from given EveEntity"""
        eve_entity = EveEntity.objects.get(id=1001)
        obj, created = MailEntity.objects.update_or_create_from_eve_entity_id(
            eve_entity.id
        )

        self.assertTrue(created)
        self.assertEqual(obj.category, MailEntity.Category.CHARACTER)
        self.assertEqual(obj.name, "Bruce Wayne")

    def test_update_or_create_from_eve_entity_id_2(self):
        """When entity already exist, update it from given EveEntity"""
        create_mail_entity(
            id=1001, category=MailEntity.Category.CHARACTER, name="John Doe"
        )

        eve_entity = EveEntity.objects.get(id=1001)
        obj, created = MailEntity.objects.update_or_create_from_eve_entity_id(
            eve_entity.id
        )

        self.assertFalse(created)
        self.assertEqual(obj.category, MailEntity.Category.CHARACTER)
        self.assertEqual(obj.name, "Bruce Wayne")

    def test_bulk_resolve_1(self):
        """Can resolve all 3 categories known by EveEntity"""
        obj_1001 = create_mail_entity(id=1001, category=MailEntity.Category.CHARACTER)
        obj_2001 = create_mail_entity(id=2001, category=MailEntity.Category.CORPORATION)
        obj_3001 = create_mail_entity(id=3001, category=MailEntity.Category.ALLIANCE)

        MailEntity.objects.bulk_update_names([obj_1001, obj_2001, obj_3001])

        self.assertEqual(obj_1001.name, "Bruce Wayne")
        self.assertEqual(obj_2001.name, "Wayne Technologies")
        self.assertEqual(obj_3001.name, "Wayne Enterprises")

    def test_bulk_resolve_2(self):
        """Will ignore categories not known to EveEntity"""

        obj_1001 = create_mail_entity(id=1001, category=MailEntity.Category.CHARACTER)
        obj_9001 = create_mail_entity(
            id=9001, category=MailEntity.Category.MAILING_LIST
        )
        obj_9002 = create_mail_entity(id=9002, category=MailEntity.Category.UNKNOWN)

        MailEntity.objects.bulk_update_names([obj_1001, obj_9001, obj_9002])

        self.assertEqual(obj_1001.name, "Bruce Wayne")
        self.assertEqual(obj_9001.name, "")
        self.assertEqual(obj_9002.name, "")

    def test_bulk_resolve_3(self):
        """When object list is empty, then no op"""

        try:
            MailEntity.objects.bulk_update_names([])
        except Exception as ex:
            self.fail(f"Unexpected exception: {ex}")

    def test_bulk_resolve_4(self):
        """When object already has a name, then update it"""
        obj_1001 = create_mail_entity(
            id=1001, category=MailEntity.Category.CHARACTER, name="John Doe"
        )

        MailEntity.objects.bulk_update_names([obj_1001])

        self.assertEqual(obj_1001.name, "Bruce Wayne")

    def test_bulk_resolve_5(self):
        """When object already has a name and respective option is chosen
        then ignore it
        """
        obj_1001 = create_mail_entity(
            id=1001, category=MailEntity.Category.CHARACTER, name="John Doe"
        )

        MailEntity.objects.bulk_update_names([obj_1001], keep_names=True)

        self.assertEqual(obj_1001.name, "John Doe")


@override_settings(CELERY_ALWAYS_EAGER=True, CELERY_EAGER_PROPAGATES_EXCEPTIONS=True)
@patch(MANAGERS_PATH + ".fetch_esi_status", spec=True)
class TestMailEntityManagerAsync(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_entities()

    def test_get_or_create_esi_async_1(self, mock_fetch_esi_status):
        """When entity already exists, return it"""
        mock_fetch_esi_status.return_value = EsiStatus(True, 99, 60)

        create_mail_entity(
            id=1234, category=MailEntity.Category.CHARACTER, name="John Doe"
        )

        obj, created = MailEntity.objects.get_or_create_esi_async(id=1234)

        self.assertFalse(created)
        self.assertEqual(obj.category, MailEntity.Category.CHARACTER)
        self.assertEqual(obj.name, "John Doe")
        self.assertFalse(mock_fetch_esi_status.called)  # was esi error status checked

    def test_get_or_create_esi_async_2(self, mock_fetch_esi_status):
        """When entity does not exist and no category specified,
        then create it asynchronously from ESI / existing EveEntity
        """
        mock_fetch_esi_status.return_value = EsiStatus(True, 99, 60)

        obj, created = MailEntity.objects.get_or_create_esi_async(id=1001)

        self.assertTrue(created)
        self.assertEqual(obj.category, MailEntity.Category.UNKNOWN)
        self.assertEqual(obj.name, "")

        obj.refresh_from_db()
        self.assertEqual(obj.category, MailEntity.Category.CHARACTER)
        self.assertEqual(obj.name, "Bruce Wayne")
        self.assertTrue(mock_fetch_esi_status.called)  # was esi error status checked

    def test_get_or_create_esi_async_3(self, mock_fetch_esi_status):
        """When entity does not exist and category is not mailing list,
        then create it synchronously from ESI / existing EveEntity
        """
        mock_fetch_esi_status.return_value = EsiStatus(True, 99, 60)

        obj, created = MailEntity.objects.get_or_create_esi_async(
            id=1001, category=MailEntity.Category.CHARACTER
        )

        self.assertTrue(created)
        self.assertEqual(obj.category, MailEntity.Category.CHARACTER)
        self.assertEqual(obj.name, "Bruce Wayne")
        self.assertFalse(mock_fetch_esi_status.called)  # was esi error status checked

    def test_update_or_create_esi_async_1(self, mock_fetch_esi_status):
        """When entity does not exist, create empty object and run task to resolve"""
        mock_fetch_esi_status.return_value = EsiStatus(True, 99, 60)

        obj, created = MailEntity.objects.update_or_create_esi_async(1001)

        self.assertTrue(created)
        self.assertEqual(obj.category, MailEntity.Category.UNKNOWN)
        self.assertEqual(obj.name, "")

        obj.refresh_from_db()
        self.assertEqual(obj.category, MailEntity.Category.CHARACTER)
        self.assertEqual(obj.name, "Bruce Wayne")

        self.assertTrue(mock_fetch_esi_status.called)  # was esi error status checked

    def test_update_or_create_esi_async_2(self, mock_fetch_esi_status):
        """When entity exists and not a mailing list, then update synchronously"""
        mock_fetch_esi_status.return_value = EsiStatus(True, 99, 60)
        create_mail_entity(
            id=1001, category=MailEntity.Category.CHARACTER, name="John Doe"
        )

        obj, created = MailEntity.objects.update_or_create_esi_async(1001)

        self.assertFalse(created)
        self.assertEqual(obj.category, MailEntity.Category.CHARACTER)
        self.assertEqual(obj.name, "Bruce Wayne")

        self.assertFalse(mock_fetch_esi_status.called)  # was esi error status checked

    def test_update_or_create_esi_async_3(self, mock_fetch_esi_status):
        """When entity exists and is a mailing list, then do nothing"""
        mock_fetch_esi_status.return_value = EsiStatus(True, 99, 60)
        create_mail_entity(
            id=9001, category=MailEntity.Category.MAILING_LIST, name="Dummy"
        )

        obj, created = MailEntity.objects.update_or_create_esi_async(9001)

        self.assertFalse(created)
        self.assertEqual(obj.category, MailEntity.Category.MAILING_LIST)
        self.assertEqual(obj.name, "Dummy")

        self.assertFalse(mock_fetch_esi_status.called)  # was esi error status checked

    def test_update_or_create_esi_async_4(self, mock_fetch_esi_status):
        """When entity does not exist and category is not a mailing list,
        then create empty object from ESI synchronously
        """
        mock_fetch_esi_status.return_value = EsiStatus(True, 99, 60)

        obj, created = MailEntity.objects.update_or_create_esi_async(
            1001, MailEntity.Category.CHARACTER
        )

        self.assertTrue(created)
        self.assertEqual(obj.category, MailEntity.Category.CHARACTER)
        self.assertEqual(obj.name, "Bruce Wayne")

        self.assertFalse(mock_fetch_esi_status.called)  # was esi error status checked


@patch(MANAGERS_PATH + ".fetch_esi_status", MagicMock(spec=fetch_esi_status))
class TestMailEntityManagerAsync2(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_entities()

    @patch(TASKS_PATH + ".update_mail_entity_esi", spec=True)
    def test_should_create_new_object_and_try_to_resolve(
        self, mock_task_update_mail_entity_esi
    ):
        # when
        obj, created = MailEntity.objects.update_or_create_esi_async(1001)
        # then
        self.assertTrue(created)
        self.assertEqual(obj.category, MailEntity.Category.UNKNOWN)
        self.assertEqual(obj.name, "")
        self.assertTrue(mock_task_update_mail_entity_esi.apply_async.called)

    @patch("memberaudit.tasks.update_mail_entity_esi", spec=True)
    def test_should_create_new_object_and_try_to_resolve_and_ignore_already_queued(
        self, mock_task_update_mail_entity_esi
    ):
        # given
        mock_task_update_mail_entity_esi.apply_async.side_effect = AlreadyQueued(10)
        # when
        obj, created = MailEntity.objects.update_or_create_esi_async(1001)
        # then
        self.assertTrue(created)
        self.assertEqual(obj.category, MailEntity.Category.UNKNOWN)
        self.assertEqual(obj.name, "")
        self.assertTrue(mock_task_update_mail_entity_esi.apply_async.called)


@patch(MANAGERS_PATH + ".esi")
@patch(MANAGERS_PATH + ".fetch_esi_status", lambda: EsiStatus(True, 99, 60))
@override_settings(CELERY_ALWAYS_EAGER=True, CELERY_EAGER_PROPAGATES_EXCEPTIONS=True)
class TestLocationManager(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()
        load_entities()
        cls.jita = EveSolarSystem.objects.get(id=30000142)
        cls.amamake = EveSolarSystem.objects.get(id=30002537)
        cls.astrahus = EveType.objects.get(id=35832)
        cls.athanor = EveType.objects.get(id=35835)
        cls.jita_trade_hub = EveType.objects.get(id=52678)
        cls.corporation_2001 = EveEntity.objects.get(id=2001)
        cls.corporation_2002 = EveEntity.objects.get(id=2002)
        cls.character = create_memberaudit_character(1001)
        cls.token = (
            cls.character.eve_character.character_ownership.user.token_set.first()
        )

    def test_can_create_structure(self, mock_esi):
        # given
        mock_esi.client = esi_client_stub

        # when
        obj, created = Location.objects.update_or_create_esi(
            id=1000000000001, token=self.token
        )

        # then
        self.assertTrue(created)
        self.assertEqual(obj.id, 1000000000001)
        self.assertEqual(obj.name, "Amamake - Test Structure Alpha")
        self.assertEqual(obj.eve_solar_system, self.amamake)
        self.assertEqual(obj.eve_type, self.astrahus)
        self.assertEqual(obj.owner, self.corporation_2001)

    def test_can_handle_incomplete_data_from_esi(self, mock_esi):
        # given
        mock_esi.client.Universe.get_universe_structures_structure_id.return_value = (
            BravadoOperationStub(
                {
                    "owner_id": None,
                    "name": "Incomplete data",
                    "solar_system_id": 30002537,
                }
            )
        )

        # when
        obj, created = Location.objects.update_or_create_esi(
            id=1000000000666, token=self.token
        )

        # then
        self.assertTrue(created)
        self.assertEqual(obj.id, 1000000000666)
        self.assertEqual(obj.name, "Incomplete data")
        self.assertEqual(obj.eve_solar_system, self.amamake)
        self.assertIsNone(obj.eve_type)
        self.assertIsNone(obj.owner)

    def test_can_update_structure(self, mock_esi):
        # given
        mock_esi.client = esi_client_stub

        # when
        obj, _ = Location.objects.update_or_create_esi(
            id=1000000000001, token=self.token
        )

        # then
        obj.name = "Not my structure"
        obj.eve_solar_system = self.jita
        obj.eve_type = self.jita_trade_hub
        obj.owner = self.corporation_2002
        obj.save()
        obj, created = Location.objects.update_or_create_esi(
            id=1000000000001, token=self.token
        )
        self.assertFalse(created)
        self.assertEqual(obj.id, 1000000000001)
        self.assertEqual(obj.name, "Amamake - Test Structure Alpha")
        self.assertEqual(obj.eve_solar_system, self.amamake)
        self.assertEqual(obj.eve_type, self.astrahus)
        self.assertEqual(obj.owner, self.corporation_2001)

    def test_does_not_update_existing_location_during_grace_period(self, mock_esi):
        # given
        mock_esi.client = esi_client_stub
        obj_existing = Location.objects.create(
            id=1000000000001,
            name="Existing Structure",
            eve_solar_system=self.jita,
            eve_type=self.jita_trade_hub,
            owner=self.corporation_2002,
        )

        # when
        obj, created = Location.objects.get_or_create_esi(
            id=1000000000001, token=self.token
        )
        # then
        self.assertFalse(created)
        self.assertEqual(obj, obj_existing)

    def test_always_update_existing_empty_locations_after_grace_period_1(
        self, mock_esi
    ):
        # given
        mock_esi.client = esi_client_stub
        Location.objects.create(id=1000000000001)

        # when
        obj, _ = Location.objects.get_or_create_esi(id=1000000000001, token=self.token)

        # then
        self.assertIsNone(obj.eve_solar_system)

    def test_always_update_existing_empty_locations_after_grace_period_2(
        self, mock_esi
    ):
        # given
        mock_esi.client = esi_client_stub
        mocked_update_at = now() - dt.timedelta(minutes=6)

        # when
        with patch(
            "django.utils.timezone.now", MagicMock(return_value=mocked_update_at)
        ):
            Location.objects.create(id=1000000000001)
            obj, _ = Location.objects.get_or_create_esi(
                id=1000000000001, token=self.token
            )

        # then
        self.assertEqual(obj.eve_solar_system, self.amamake)

    @patch(MANAGERS_PATH + ".MEMBERAUDIT_LOCATION_STALE_HOURS", 24)
    def test_always_update_existing_locations_which_are_stale(self, mock_esi):
        # given
        mock_esi.client = esi_client_stub
        mocked_update_at = now() - dt.timedelta(hours=25)
        with patch(
            "django.utils.timezone.now", MagicMock(return_value=mocked_update_at)
        ):
            Location.objects.create(
                id=1000000000001,
                name="Existing Structure",
                eve_solar_system=self.jita,
                eve_type=self.jita_trade_hub,
                owner=self.corporation_2002,
            )

        # when
        obj, created = Location.objects.get_or_create_esi(
            id=1000000000001, token=self.token
        )

        # then
        self.assertFalse(created)
        self.assertEqual(obj.eve_solar_system, self.amamake)

    def test_propagates_http_error_on_structure_create(self, mock_esi):
        # given
        mock_esi.client = esi_client_stub

        # when/Then
        with self.assertRaises(HTTPNotFound):
            Location.objects.update_or_create_esi(id=1000000000099, token=self.token)

    def test_always_creates_empty_location_for_invalid_ids(self, mock_esi):
        # given
        mock_esi.client = esi_client_stub

        # when
        obj, created = Location.objects.update_or_create_esi(
            id=80000000, token=self.token
        )

        # then
        self.assertTrue(created)
        self.assertTrue(obj.is_empty)

    def test_propagates_exceptions_on_structure_create(self, mock_esi):
        mock_esi.client.Universe.get_universe_structures_structure_id.side_effect = (
            RuntimeError
        )

        with self.assertRaises(RuntimeError):
            Location.objects.update_or_create_esi(id=1000000000099, token=self.token)

    def test_can_create_empty_location_on_access_error_1(self, mock_esi):
        mock_esi.client.Universe.get_universe_structures_structure_id.side_effect = (
            HTTPForbidden(response=BravadoResponseStub(403, "Test exception"))
        )

        obj, created = Location.objects.update_or_create_esi(
            id=1000000000099, token=self.token
        )
        self.assertTrue(created)
        self.assertEqual(obj.id, 1000000000099)

    def test_can_create_empty_location_on_access_error_2(self, mock_esi):
        mock_esi.client.Universe.get_universe_structures_structure_id.side_effect = (
            HTTPUnauthorized(response=BravadoResponseStub(401, "Test exception"))
        )

        obj, created = Location.objects.update_or_create_esi(
            id=1000000000099, token=self.token
        )
        self.assertTrue(created)
        self.assertEqual(obj.id, 1000000000099)

    def test_does_not_creates_empty_location_on_access_errors_if_requested(
        self, mock_esi
    ):
        mock_esi.client.Universe.get_universe_structures_structure_id.side_effect = (
            RuntimeError
        )
        with self.assertRaises(RuntimeError):
            Location.objects.update_or_create_esi(id=1000000000099, token=self.token)

    def test_records_esi_error_on_access_error(self, mock_esi):
        mock_esi.client.Universe.get_universe_structures_structure_id.side_effect = (
            HTTPForbidden(
                response=BravadoResponseStub(
                    403,
                    "Test exception",
                    headers={
                        "X-Esi-Error-Limit-Remain": "40",
                        "X-Esi-Error-Limit-Reset": "30",
                    },
                )
            )
        )

        _, created = Location.objects.update_or_create_esi(
            id=1000000000099, token=self.token
        )
        self.assertTrue(created)

    def test_should_raise_value_error_when_token_is_needed_but_not_passed(
        self, mock_esi
    ):
        # given
        mock_esi.client = esi_client_stub

        # when/then
        with self.assertRaises(ValueError):
            Location.objects.get_or_create_esi(id=1000000000099, token=None)

    # stations

    def test_can_create_station(self, mock_esi):
        mock_esi.client = esi_client_stub

        obj, created = Location.objects.update_or_create_esi(
            id=60003760, token=self.token
        )
        self.assertTrue(created)
        self.assertEqual(obj.id, 60003760)
        self.assertEqual(obj.name, "Jita IV - Moon 4 - Caldari Navy Assembly Plant")
        self.assertEqual(obj.eve_solar_system, self.jita)
        self.assertEqual(obj.eve_type, self.jita_trade_hub)
        self.assertEqual(obj.owner, self.corporation_2002)

    def test_can_update_station(self, mock_esi):
        mock_esi.client = esi_client_stub

        obj, created = Location.objects.update_or_create_esi(
            id=60003760, token=self.token
        )
        obj.name = "Not my station"
        obj.eve_solar_system = self.amamake
        obj.eve_type = self.astrahus
        obj.owner = self.corporation_2001
        obj.save()

        obj, created = Location.objects.update_or_create_esi(
            id=60003760, token=self.token
        )
        self.assertFalse(created)
        self.assertEqual(obj.id, 60003760)
        self.assertEqual(obj.name, "Jita IV - Moon 4 - Caldari Navy Assembly Plant")
        self.assertEqual(obj.eve_solar_system, self.jita)
        self.assertEqual(obj.eve_type, self.jita_trade_hub)
        self.assertEqual(obj.owner, self.corporation_2002)

    def test_propagates_http_error_on_station_create(self, mock_esi):
        mock_esi.client = esi_client_stub

        with self.assertRaises(HTTPNotFound):
            Location.objects.update_or_create_esi(id=63999999, token=self.token)

    # Solar System

    def test_can_create_solar_system(self, mock_esi):
        mock_esi.client = esi_client_stub

        obj, created = Location.objects.update_or_create_esi(
            id=30002537, token=self.token
        )
        self.assertTrue(created)
        self.assertEqual(obj.id, 30002537)
        self.assertEqual(obj.name, "Amamake")
        self.assertEqual(obj.eve_solar_system, self.amamake)
        self.assertEqual(obj.eve_type, EveType.objects.get(id=EveTypeId.SOLAR_SYSTEM))
        self.assertIsNone(obj.owner)

    # Asset Safety

    def test_can_create_asset_safety(self, mock_esi):
        # given
        mock_esi.client = esi_client_stub
        # when
        obj, created = Location.objects.update_or_create_esi(id=2004, token=self.token)
        # then
        self.assertTrue(created)
        self.assertEqual(obj.id, 2004)
        self.assertEqual(obj.name, "ASSET SAFETY")
        self.assertIsNone(obj.eve_solar_system)
        self.assertIsNone(obj.owner)
        self.assertEqual(
            obj.eve_type, EveType.objects.get(id=EveTypeId.ASSET_SAFETY_WRAP)
        )

    # Unknown location placeholder

    def test_can_create_unknown_location(self, mock_esi):
        # given
        mock_esi.client = esi_client_stub
        # when
        obj, created = Location.objects.update_or_create_esi(id=888, token=self.token)
        # then
        self.assertTrue(created)
        self.assertEqual(obj.id, 888)
        self.assertEqual(obj.name, "Location unknown")
        self.assertIsNone(obj.eve_solar_system)
        self.assertIsNone(obj.owner)
        self.assertEqual(obj.eve_type_id, EveTypeId.SOLAR_SYSTEM)

    def test_should_create_obj_from_solar_system(self, mock_esi):
        # when
        obj, created = Location.objects.get_or_create_from_eve_solar_system(self.jita)
        # then
        self.assertTrue(created)
        self.assertEqual(obj.id, self.jita.id)
        self.assertEqual(obj.name, self.jita.name)
        self.assertEqual(obj.eve_solar_system, self.jita)
        self.assertEqual(obj.eve_type_id, EveTypeId.SOLAR_SYSTEM)

    def test_should_get_existing_obj_from_solar_system(self, mock_esi):
        # given
        create_location(
            id=self.jita.id,
            eve_solar_system=self.jita,
            name=self.jita.name,
            eve_type_id=EveTypeId.SOLAR_SYSTEM,
        )
        # when
        obj, created = Location.objects.get_or_create_from_eve_solar_system(self.jita)
        # then
        self.assertFalse(created)
        self.assertEqual(obj.id, self.jita.id)
        self.assertEqual(obj.name, self.jita.name)
        self.assertEqual(obj.eve_solar_system, self.jita)
        self.assertEqual(obj.eve_type_id, EveTypeId.SOLAR_SYSTEM)

    def test_should_create_unknown_location_object_when_it_does_not_exist(
        self, mock_esi
    ):
        # when
        obj, created = Location.objects.get_or_create_unknown_location()
        # then
        self.assertTrue(created)
        self.assertTrue(obj.is_unknown_location)

    def test_should_return_existing_unknown_location_object(self, mock_esi):
        # given
        Location.objects.get_or_create_unknown_location()
        # when
        obj, created = Location.objects.get_or_create_unknown_location()
        # then
        self.assertFalse(created)
        self.assertTrue(obj.is_unknown_location)


@patch(MANAGERS_PATH + ".esi")
@patch(MANAGERS_PATH + ".LocationManager.get_or_create_esi_async")
class TestLocationManagerPreload(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()
        load_entities()
        cls.token = MagicMock(spec=Token)

    def test_can_preload_missing_locations(
        self, mock_get_or_create_esi_async, mock_esi
    ):
        # given
        mock_esi.client = esi_client_stub
        Location.objects.update_or_create_esi(id=60003760, token=self.token)
        # when
        result = Location.objects.create_missing_esi([60003760, 30002537], self.token)
        # then
        self.assertEqual(mock_get_or_create_esi_async.call_count, 1)
        _, kwargs = mock_get_or_create_esi_async.call_args
        self.assertEqual(kwargs["id"], 30002537)
        self.assertSetEqual(result, {60003760, 30002537})

    def test_can_do_nothing_when_all_locations_found(
        self, mock_get_or_create_esi_async, mock_esi
    ):
        # given
        mock_esi.client = esi_client_stub
        Location.objects.update_or_create_esi(id=60003760, token=self.token)
        Location.objects.update_or_create_esi(id=30002537, token=self.token)
        # when
        result = Location.objects.create_missing_esi([60003760, 30002537], self.token)
        # then
        self.assertEqual(mock_get_or_create_esi_async.call_count, 0)
        self.assertSetEqual(result, {60003760, 30002537})

    def test_can_do_nothing_when_no_ids_provided(
        self, mock_get_or_create_esi_async, mock_esi
    ):
        # given
        mock_esi.client = esi_client_stub
        # when
        result = Location.objects.create_missing_esi([], self.token)
        # then
        self.assertEqual(mock_get_or_create_esi_async.call_count, 0)
        self.assertSetEqual(result, set())


@override_settings(CELERY_ALWAYS_EAGER=True, CELERY_EAGER_PROPAGATES_EXCEPTIONS=True)
class TestLocationManagerAsync(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()
        load_entities()
        cls.jita = EveSolarSystem.objects.get(id=30000142)
        cls.amamake = EveSolarSystem.objects.get(id=30002537)
        cls.astrahus = EveType.objects.get(id=35832)
        cls.athanor = EveType.objects.get(id=35835)
        cls.jita_trade_hub = EveType.objects.get(id=52678)
        cls.corporation_2001 = EveEntity.objects.get(id=2001)
        cls.corporation_2002 = EveEntity.objects.get(id=2002)
        cls.character = create_memberaudit_character(1001)
        cls.token = (
            cls.character.eve_character.character_ownership.user.token_set.first()
        )

    @patch(MANAGERS_PATH + ".esi")
    @patch(MANAGERS_PATH + ".fetch_esi_status", spec=True)
    def test_can_create_structure_async(self, mock_fetch_esi_status, mock_esi):
        # given
        mock_fetch_esi_status.return_value = EsiStatus(True, 99, 60)
        mock_esi.client = esi_client_stub
        # when
        obj, created = Location.objects.update_or_create_esi_async(
            id=1000000000001, token=self.token
        )
        # then
        self.assertTrue(created)
        self.assertEqual(obj.id, 1000000000001)
        self.assertIsNone(obj.eve_solar_system)
        self.assertIsNone(obj.eve_type)
        obj.refresh_from_db()
        self.assertEqual(obj.name, "Amamake - Test Structure Alpha")
        self.assertEqual(obj.eve_solar_system, self.amamake)
        self.assertEqual(obj.eve_type, self.astrahus)
        self.assertEqual(obj.owner, self.corporation_2001)
        self.assertTrue(mock_fetch_esi_status.called)  # proofs task was called

    @patch(MANAGERS_PATH + ".fetch_esi_status", MagicMock(spec=fetch_esi_status))
    @patch(TASKS_PATH + ".update_structure_esi", spec=True)
    def test_should_create_location_and_ignore_already_queued(
        self, mock_task_update_structure_esi
    ):
        # given
        mock_task_update_structure_esi.apply_async.side_effect = AlreadyQueued(10)
        # when
        obj, created = Location.objects.update_or_create_esi_async(
            id=1000000000001, token=self.token
        )
        # then
        self.assertTrue(created)
        self.assertEqual(obj.id, 1000000000001)
        self.assertIsNone(obj.eve_solar_system)
        self.assertIsNone(obj.eve_type)
        self.assertTrue(mock_task_update_structure_esi.apply_async.called)


@patch(MANAGERS_PATH + ".esi")
class TestCharacterMailingLists(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_entities()
        cls.character = create_memberaudit_character(1001)

    def test_update_mailing_lists_1(self, mock_esi):
        """can create new mailing lists from scratch"""
        mock_esi.client = esi_client_stub

        self.character.update_mailing_lists()

        self.assertSetEqual(
            set(MailEntity.objects.values_list("id", flat=True)), {9001, 9002}
        )
        self.assertSetEqual(
            set(self.character.mailing_lists.values_list("id", flat=True)),
            {9001, 9002},
        )

        obj = MailEntity.objects.get(id=9001)
        self.assertEqual(obj.name, "Dummy 1")

        obj = MailEntity.objects.get(id=9002)
        self.assertEqual(obj.name, "Dummy 2")

    def test_update_mailing_lists_2(self, mock_esi):
        """does not remove obsolete mailing lists"""
        mock_esi.client = esi_client_stub
        create_mail_entity(
            id=5, category=MailEntity.Category.MAILING_LIST, name="Obsolete"
        )

        self.character.update_mailing_lists()

        self.assertSetEqual(
            set(MailEntity.objects.values_list("id", flat=True)), {9001, 9002, 5}
        )
        self.assertSetEqual(
            set(self.character.mailing_lists.values_list("id", flat=True)),
            {9001, 9002},
        )

    def test_update_mailing_lists_3(self, mock_esi):
        """updates existing mailing lists"""
        mock_esi.client = esi_client_stub
        create_mail_entity(
            id=9001, category=MailEntity.Category.MAILING_LIST, name="Update me"
        )

        self.character.update_mailing_lists()

        self.assertSetEqual(
            set(MailEntity.objects.values_list("id", flat=True)), {9001, 9002}
        )
        self.assertSetEqual(
            set(self.character.mailing_lists.values_list("id", flat=True)),
            {9001, 9002},
        )
        obj = MailEntity.objects.get(id=9001)
        self.assertEqual(obj.name, "Dummy 1")

    def test_update_mailing_lists_4(self, mock_esi):
        """when data from ESI has not changed, then skip update"""
        mock_esi.client = esi_client_stub

        self.character.update_mailing_lists()
        obj = MailEntity.objects.get(id=9001)
        obj.name = "Extravaganza"
        obj.save()
        self.character.mailing_lists.clear()

        self.character.update_mailing_lists()
        obj = MailEntity.objects.get(id=9001)
        self.assertEqual(obj.name, "Extravaganza")
        self.assertSetEqual(
            set(self.character.mailing_lists.values_list("id", flat=True)), set()
        )

    def test_update_mailing_lists_5(self, mock_esi):
        """when data from ESI has not changed and update is forced, then do update"""
        mock_esi.client = esi_client_stub

        self.character.update_mailing_lists()
        obj = MailEntity.objects.get(id=9001)
        obj.name = "Extravaganza"
        obj.save()

        self.character.update_mailing_lists(force_update=True)
        obj = MailEntity.objects.get(id=9001)
        self.assertEqual(obj.name, "Dummy 1")


class TestSkillSetManager(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()
        load_entities()
        cls.fitting = create_fitting(name="My fitting")

    def test_should_create_new_skill_set_from_fitting(self):
        # when
        skill_set, created = SkillSet.objects.update_or_create_from_fitting(
            fitting=self.fitting
        )
        # then
        self.assertTrue(created)
        self.assertEqual(skill_set.name, "My fitting")
        self.assertEqual(skill_set.ship_type.name, "Tristan")
        skills_str = {skill.required_skill_str for skill in skill_set.skills.all()}
        self.assertSetEqual(
            skills_str,
            {
                "Small Autocannon Specialization I",
                "Gunnery II",
                "Weapon Upgrades IV",
                "Light Drone Operation V",
                "Small Projectile Turret V",
                "Gallente Frigate I",
                "Propulsion Jamming II",
                "Drones V",
                "Amarr Drone Specialization I",
            },
        )

    def test_should_create_new_skill_set_from_fitting_and_assign_to_group(self):
        # given
        skill_set_group = create_skill_set_group()
        # when
        skill_set, created = SkillSet.objects.update_or_create_from_fitting(
            fitting=self.fitting, skill_set_group=skill_set_group
        )
        # then
        self.assertTrue(created)
        self.assertIn(skill_set, skill_set_group.skill_sets.all())

    def test_should_create_new_skill_set_from_skill_plan(self):
        # given
        skills = [
            create_skill(
                eve_type=EveType.objects.get(name="Small Autocannon Specialization"),
                level=1,
            ),
            create_skill(
                eve_type=EveType.objects.get(name="Light Drone Operation"),
                level=5,
            ),
        ]
        skill_plan = create_skill_plan(name="My skill plan", skills=skills)
        # when
        skill_set, created = SkillSet.objects.update_or_create_from_skill_plan(
            skill_plan=skill_plan
        )
        # then
        self.assertTrue(created)
        self.assertEqual(skill_set.name, "My skill plan")
        skills_str = {skill.required_skill_str for skill in skill_set.skills.all()}
        self.assertSetEqual(
            skills_str,
            {"Small Autocannon Specialization I", "Light Drone Operation V"},
        )

    def test_should_create_new_skill_set_from_skill_plan_and_assign_to_group(self):
        # given
        # given
        skills = [
            create_skill(
                eve_type=EveType.objects.get(name="Small Autocannon Specialization"),
                level=1,
            ),
            create_skill(
                eve_type=EveType.objects.get(name="Light Drone Operation"),
                level=5,
            ),
        ]
        skill_plan = create_skill_plan(name="My skill plan", skills=skills)
        skill_set_group = create_skill_set_group()
        # when
        skill_set, created = SkillSet.objects.update_or_create_from_skill_plan(
            skill_plan=skill_plan, skill_set_group=skill_set_group
        )
        # then
        self.assertTrue(created)
        self.assertIn(skill_set, skill_set_group.skill_sets.all())



"""Managers for character section models (3/3)."""
# pylint: disable=missing-class-docstring

from __future__ import annotations

from typing import TYPE_CHECKING, List, Set

from django.db import models, transaction
from django.db.models import ExpressionWrapper, F
from django.utils.html import strip_tags
from esi.models import Token
from eveuniverse.models import EveEntity, EvePlanet, EveSolarSystem, EveType

from allianceauth.services.hooks import get_extension_logger
from app_utils.logging import LoggerAddTag

from memberaudit import __title__
from memberaudit.app_settings import MEMBERAUDIT_BULK_METHODS_BATCH_SIZE
from memberaudit.decorators import fetch_token_for_character
from memberaudit.helpers import (
    UpdateSectionResult,
    data_retention_cutoff,
    eve_entity_ids_from_objs,
)
from memberaudit.managers._common import GenericUpdateComplexObjMixin
from memberaudit.providers import esi
from memberaudit.utils import (
    get_or_create_esi_or_none,
    get_or_create_or_none,
    get_or_none,
)

from ._common import GenericUpdateSimpleObjMixin

if TYPE_CHECKING:
    from memberaudit.models import Character, CharacterSkillqueueEntry

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


class CharacterMiningLedgerEntryQueryset(models.QuerySet):
    def annotate_pricing(self) -> models.QuerySet:
        """Annotate price and total columns."""
        return (
            self.select_related("eve_type__market_price")
            .annotate(price=F("eve_type__market_price__average_price"))
            .annotate(
                total=ExpressionWrapper(
                    F("eve_type__market_price__average_price") * F("quantity"),
                    output_field=models.FloatField(),
                ),
            )
        )


class CharacterMiningLedgerEntryManagerBase(models.Manager):
    def update_or_create_esi(
        self, character: Character, force_update: bool = False
    ) -> UpdateSectionResult:
        """Update or create mining ledger for a character from ESI."""

        return character.update_section_if_changed(
            section=character.UpdateSection.MINING_LEDGER,
            fetch_func=self._fetch_data_from_esi,
            store_func=self._update_or_create_objs,
            force_update=force_update,
        )

    @fetch_token_for_character("esi-industry.read_character_mining.v1")
    def _fetch_data_from_esi(self, character: Character, token: Token):
        logger.info("%s: Fetching mining ledger from ESI", character)
        entries = esi.client.Industry.get_characters_character_id_mining(
            character_id=character.eve_character.character_id,
            token=token.valid_access_token(),
        ).results()
        return entries

    def _update_or_create_objs(self, character: Character, entries):
        # preload solar systems
        solar_system_ids = {entry["solar_system_id"] for entry in entries}
        for solar_system_id in solar_system_ids:
            EveSolarSystem.objects.get_or_create_esi(id=solar_system_id)

        # preload eve types
        type_ids = {entry["type_id"] for entry in entries}
        for type_id in type_ids:
            EveType.objects.get_or_create_esi(id=type_id)

        # store entries
        for entry in entries:
            self.update_or_create(
                character=character,
                date=entry["date"],
                eve_solar_system_id=entry["solar_system_id"],
                eve_type_id=entry["type_id"],
                defaults={"quantity": entry["quantity"]},
            )


CharacterMiningLedgerEntryManager = CharacterMiningLedgerEntryManagerBase.from_queryset(
    CharacterMiningLedgerEntryQueryset
)


class CharacterOnlineStatusManager(models.Manager):
    def update_or_create_esi(
        self, character: Character, force_update: bool = False
    ) -> UpdateSectionResult:
        """Update or create online status for a character from ESI."""
        return character.update_section_if_changed(
            section=character.UpdateSection.ONLINE_STATUS,
            fetch_func=self._fetch_data_from_esi,
            store_func=self._update_or_create_objs,
            force_update=force_update,
        )

    @fetch_token_for_character("esi-location.read_online.v1")
    def _fetch_data_from_esi(self, character: Character, token):
        logger.info("%s: Fetching online status from ESI", character)
        online_info = esi.client.Location.get_characters_character_id_online(
            character_id=character.eve_character.character_id,
            token=token.valid_access_token(),
        ).results()

        return online_info

    def _update_or_create_objs(self, character: Character, online_info):
        self.update_or_create(
            character=character,
            defaults={
                "last_login": online_info.get("last_login"),
                "last_logout": online_info.get("last_logout"),
                "logins": online_info.get("logins"),
            },
        )


class CharacterRoleManager(models.Manager):
    def update_or_create_esi(
        self, character: Character, force_update: bool = False
    ) -> UpdateSectionResult:
        """Update or create roles for a character from ESI."""

        return character.update_section_if_changed(
            section=character.UpdateSection.ROLES,
            fetch_func=self._fetch_data_from_esi,
            store_func=self._update_or_create_objs,
            force_update=force_update,
        )

    @fetch_token_for_character("esi-characters.read_corporation_roles.v1")
    def _fetch_data_from_esi(self, character: Character, token: Token) -> dict:
        """Update the character's roles"""

        logger.info("%s: Fetching roles from ESI", character)
        roles_data = esi.client.Character.get_characters_character_id_roles(
            character_id=character.eve_character.character_id,
            token=token.valid_access_token(),
        ).results()
        return roles_data

    @transaction.atomic()
    def _update_or_create_objs(self, character: Character, roles_data: dict):
        from memberaudit.models import CharacterRole

        Role = CharacterRole.Role
        roles_map = {
            "Account_Take_1": Role.ACCOUNT_TAKE_1,
            "Account_Take_2": Role.ACCOUNT_TAKE_2,
            "Account_Take_3": Role.ACCOUNT_TAKE_3,
            "Account_Take_4": Role.ACCOUNT_TAKE_4,
            "Account_Take_5": Role.ACCOUNT_TAKE_5,
            "Account_Take_6": Role.ACCOUNT_TAKE_6,
            "Account_Take_7": Role.ACCOUNT_TAKE_7,
            "Accountant": Role.ACCOUNTANT,
            "Auditor": Role.AUDITOR,
            "Communications_Officer": Role.COMMUNICATIONS_OFFICER,
            "Config_Equipment": Role.CONFIG_EQUIPMENT,
            "Config_Starbase_Equipment": Role.CONFIG_STARBASE_EQUIPMENT,
            "Container_Take_1": Role.CONTAINER_TAKE_1,
            "Container_Take_2": Role.CONTAINER_TAKE_2,
            "Container_Take_3": Role.CONTAINER_TAKE_3,
            "Container_Take_4": Role.CONTAINER_TAKE_4,
            "Container_Take_5": Role.CONTAINER_TAKE_5,
            "Container_Take_6": Role.CONTAINER_TAKE_6,
            "Container_Take_7": Role.CONTAINER_TAKE_7,
            "Contract_Manager": Role.CONTRACT_MANAGER,
            "Diplomat": Role.DIPLOMAT,
            "Director": Role.DIRECTOR,
            "Factory_Manager": Role.FACTORY_MANAGER,
            "Fitting_Manager": Role.FITTING_MANAGER,
            "Hangar_Query_1": Role.HANGAR_QUERY_1,
            "Hangar_Query_2": Role.HANGAR_QUERY_2,
            "Hangar_Query_3": Role.HANGAR_QUERY_3,
            "Hangar_Query_4": Role.HANGAR_QUERY_4,
            "Hangar_Query_5": Role.HANGAR_QUERY_5,
            "Hangar_Query_6": Role.HANGAR_QUERY_6,
            "Hangar_Query_7": Role.HANGAR_QUERY_7,
            "Hangar_Take_1": Role.HANGAR_TAKE_1,
            "Hangar_Take_2": Role.HANGAR_TAKE_2,
            "Hangar_Take_3": Role.HANGAR_TAKE_3,
            "Hangar_Take_4": Role.HANGAR_TAKE_4,
            "Hangar_Take_5": Role.HANGAR_TAKE_5,
            "Hangar_Take_6": Role.HANGAR_TAKE_6,
            "Hangar_Take_7": Role.HANGAR_TAKE_7,
            "Junior_Accountant": Role.JUNIOR_ACCOUNTANT,
            "Personnel_Manager": Role.PERSONNEL_MANAGER,
            "Rent_Factory_Facility": Role.RENT_FACTORY_FACILITY,
            "Rent_Office": Role.RENT_OFFICE,
            "Rent_Research_Facility": Role.RENT_RESEARCH_FACILITY,
            "Security_Officer": Role.SECURITY_OFFICER,
            "Skill_Plan_Manager": Role.SKILL_PLAN_MANAGER,
            "Starbase_Defense_Operator": Role.STARBASE_DEFENSE_OPERATOR,
            "Starbase_Fuel_Technician": Role.STARBASE_FUEL_TECHNICIAN,
            "Station_Manager": Role.STATION_MANAGER,
            "Trader": Role.TRADER,
        }
        Location = CharacterRole.Location
        location_map = {
            "roles": Location.UNIVERSAL,
            "roles_at_base": Location.BASE,
            "roles_at_hq": Location.HQ,
            "roles_at_other": Location.OTHER,
        }
        to_remove = list(
            self.filter(character=character).values_list("location", "role")
        )
        to_add = []
        for location_name, roles in roles_data.items():
            location = location_map[location_name]
            for role_name in roles:
                role = roles_map[role_name]
                if (location, role) in to_remove:
                    # if we already have the role, don't remove it
                    to_remove.remove((location, role))
                else:
                    # if we don't have the role, prepare to add it
                    to_add.append(
                        self.model(character=character, role=role, location=location)
                    )
        if to_add:
            self.bulk_create(to_add)
            logger.info("%s: Added %d new roles", character, len(to_add))

        if to_remove:
            for location, role in to_remove:
                self.filter(character=character, location=location, role=role).delete()
            logger.info("%s: Removed %d obsolete roles", character, len(to_add))


class CharacterPlanetManager(GenericUpdateComplexObjMixin, models.Manager):
    def update_or_create_esi(
        self, character: Character, force_update: bool = False
    ) -> UpdateSectionResult:
        """Update or create planets for a character from ESI."""
        return character.update_section_if_changed(
            section=character.UpdateSection.PLANETS,
            fetch_func=self._fetch_data_from_esi,
            store_func=self._update_or_create_objs,
            force_update=force_update,
        )

    @fetch_token_for_character("esi-planets.manage_planets.v1")
    def _fetch_data_from_esi(self, character: Character, token: Token) -> List[dict]:
        logger.info("%s: Fetching planets from ESI", character)
        planets_data = (
            esi.client.Planetary_Interaction.get_characters_character_id_planets(
                character_id=character.eve_character.character_id,
                token=token.valid_access_token(),
            ).results()
        )

        return planets_data

    def _update_or_create_objs(
        self, character: Character, esi_data: List[dict]
    ) -> Set[int]:
        def make_obj_from_esi_entry(character: Character, entry: dict):
            eve_planet = EvePlanet.objects.get_or_create_esi(id=entry["planet_id"])[0]
            obj = self.model(
                character=character,
                eve_planet=eve_planet,
                num_pins=entry["num_pins"],
                upgrade_level=entry["upgrade_level"],
                last_update_at=entry["last_update"],
            )
            return obj

        self._update_or_create_objs_generic(
            character,
            esi_data,
            model_key_field="eve_planet_id",
            fields_for_update=("num_pins", "upgrade_level", "last_update_at"),
            make_obj_from_esi_entry=make_obj_from_esi_entry,
        )


class CharacterShipManager(models.Manager):
    def update_or_create_esi(
        self, character: Character, force_update: bool = False
    ) -> UpdateSectionResult:
        """Update or create ship for a character from ESI."""
        return character.update_section_if_changed(
            section=character.UpdateSection.SHIP,
            fetch_func=self._fetch_data_from_esi,
            store_func=self._update_or_create_objs,
            force_update=force_update,
        )

    @fetch_token_for_character("esi-location.read_ship_type.v1")
    def _fetch_data_from_esi(self, character: Character, token: Token):
        logger.info("%s: Fetching ship from ESI", character)
        ship_info = esi.client.Location.get_characters_character_id_ship(
            character_id=character.eve_character.character_id,
            token=token.valid_access_token(),
        ).results()
        return ship_info

    def _update_or_create_objs(self, character: Character, ship_info):
        ship_type_id = ship_info.get("ship_type_id")
        if not ship_type_id:
            self.filter(character=character).delete()
            return

        eve_type, _ = EveType.objects.get_or_create_esi(id=ship_type_id)
        self.update_or_create(
            character=character,
            defaults={
                "item_id": ship_info["ship_item_id"],
                "eve_type": eve_type,
                "name": ship_info["ship_name"],
            },
        )


class CharacterSkillqueueEntryManager(models.Manager):
    def update_or_create_esi(
        self, character: Character, force_update: bool = False
    ) -> UpdateSectionResult:
        """Update or create skills queue for a character from ESI."""
        return character.update_section_if_changed(
            section=character.UpdateSection.SKILL_QUEUE,
            fetch_func=self._fetch_data_from_esi,
            store_func=self._update_or_create_objs,
            force_update=force_update,
        )

    @fetch_token_for_character("esi-skills.read_skillqueue.v1")
    def _fetch_data_from_esi(self, character: Character, token) -> List[dict]:
        logger.info("%s: Fetching skill queue from ESI", character)
        skillqueue = esi.client.Skills.get_characters_character_id_skillqueue(
            character_id=character.eve_character.character_id,
            token=token.valid_access_token(),
        ).results()

        return skillqueue

    def _update_or_create_objs(self, character: Character, skillqueue: List[dict]):
        entries = self._compile_objs(character, skillqueue)
        self._write_objs(character, entries)

    def _compile_objs(self, character: Character, skillqueue: List[dict]):
        if not skillqueue:
            return []

        entries = [
            self.model(
                character=character,
                eve_type=get_or_create_esi_or_none("skill_id", entry, EveType),
                finish_date=entry.get("finish_date"),
                finished_level=entry.get("finished_level"),
                level_end_sp=entry.get("level_end_sp"),
                level_start_sp=entry.get("level_start_sp"),
                queue_position=entry.get("queue_position"),
                start_date=entry.get("start_date"),
                training_start_sp=entry.get("training_start_sp"),
            )
            for entry in skillqueue
        ]
        return entries

    @transaction.atomic()
    def _write_objs(
        self, character: Character, entries: List[CharacterSkillqueueEntry]
    ):
        self.filter(character=character).delete()
        if not entries:
            logger.info("%s: Skill queue is empty", character)
            return

        self.bulk_create(entries, batch_size=MEMBERAUDIT_BULK_METHODS_BATCH_SIZE)
        logger.info("%s: Updated skill queue of size %s", character, len(entries))


class CharacterSkillManager(models.Manager):
    def update_or_create_esi(
        self, character: Character, force_update: bool = False
    ) -> UpdateSectionResult:
        """Update or create skills for a character from ESI."""
        return character.update_section_if_changed(
            section=character.UpdateSection.SKILLS,
            fetch_func=self._fetch_data_from_esi,
            store_func=self._update_or_create_objs,
            force_update=force_update,
        )

    @fetch_token_for_character("esi-skills.read_skills.v1")
    def _fetch_data_from_esi(self, character: Character, token: Token) -> dict:
        logger.info("%s: Fetching skills from ESI", character)
        skills_info = esi.client.Skills.get_characters_character_id_skills(
            character_id=character.eve_character.character_id,
            token=token.valid_access_token(),
        ).results()
        return skills_info

    def _preload_types(self, skills_list: dict):
        if skills_list:
            incoming_ids = set(skills_list.keys())
            existing_ids = set(self.values_list("eve_type_id", flat=True))
            new_ids = incoming_ids.difference(existing_ids)
            EveType.objects.bulk_get_or_create_esi(ids=list(new_ids))

    def _update_or_create_objs(self, character: Character, skills_info: dict):
        from memberaudit.models import CharacterSkillpoints

        CharacterSkillpoints.objects.update_or_create(
            character=character,
            defaults={
                "total": skills_info.get("total_sp"),
                "unallocated": skills_info.get("unallocated_sp"),
            },
        )

        skills_list = {
            obj["skill_id"]: obj
            for obj in skills_info.get("skills", [])
            if "skill_id" in obj
        }
        self._preload_types(skills_list)

        with transaction.atomic():
            incoming_ids = set(skills_list.keys())
            existing_ids = set(
                self.filter(character=character).values_list("eve_type_id", flat=True)
            )
            obsolete_ids = existing_ids.difference(incoming_ids)
            if obsolete_ids:
                logger.info(
                    "%s: Removing %s obsolete skills", character, len(obsolete_ids)
                )
                self.filter(character=character, eve_type_id__in=obsolete_ids).delete()

            create_ids = None
            update_ids = None
            if skills_list:
                create_ids = incoming_ids.difference(existing_ids)
                if create_ids:
                    self._create_from_dict(
                        character=character,
                        skills_list=skills_list,
                        create_ids=create_ids,
                    )

                update_ids = incoming_ids.difference(create_ids)
                if update_ids:
                    self._update_from_dict(
                        character=character,
                        skills_list=skills_list,
                        update_ids=update_ids,
                    )

            if not obsolete_ids and not create_ids and not update_ids:
                logger.info("%s: Skills have not changed", character)

    def _create_from_dict(
        self, character: Character, skills_list: dict, create_ids: set
    ):
        logger.info("%s: Storing %s new skills", character, len(create_ids))
        skills = [
            self.model(
                character=character,
                eve_type=EveType.objects.get(id=skill_info.get("skill_id")),
                active_skill_level=skill_info.get("active_skill_level"),
                skillpoints_in_skill=skill_info.get("skillpoints_in_skill"),
                trained_skill_level=skill_info.get("trained_skill_level"),
            )
            for skill_id, skill_info in skills_list.items()
            if skill_id in create_ids
        ]
        self.bulk_create(skills, batch_size=MEMBERAUDIT_BULK_METHODS_BATCH_SIZE)

    def _update_from_dict(
        self, character: Character, skills_list: dict, update_ids: set
    ):
        logger.info("%s: Updating %s skills", character, len(update_ids))
        update_pks = list(
            self.filter(character=character, eve_type_id__in=update_ids).values_list(
                "pk", flat=True
            )
        )
        skills = self.in_bulk(update_pks)
        for skill in skills.values():
            skill_info = skills_list.get(skill.eve_type_id)
            if skill_info:
                skill.active_skill_level = skill_info.get("active_skill_level")
                skill.skillpoints_in_skill = skill_info.get("skillpoints_in_skill")
                skill.trained_skill_level = skill_info.get("trained_skill_level")

        self.bulk_update(
            skills.values(),
            fields=[
                "active_skill_level",
                "skillpoints_in_skill",
                "trained_skill_level",
            ],
            batch_size=MEMBERAUDIT_BULK_METHODS_BATCH_SIZE,
        )

    def find_active_skill_level(self, skill_id: int) -> int:
        """Return active skill level for a skill or 0 if not found."""
        try:
            skill = self.get(eve_type_id=skill_id)
        except self.model.DoesNotExist:
            return 0
        return skill.active_skill_level


class CharacterSkillSetCheckManager(models.Manager):
    # TODO: Replace delete & create with update
    @transaction.atomic()
    def update_for_character(self, character: Character) -> UpdateSectionResult:
        """Update or create skill sets for a character."""
        from memberaudit.models import SkillSet

        character_skills = {
            obj["eve_type_id"]: obj["active_skill_level"]
            for obj in character.skills.values("eve_type_id", "active_skill_level")
        }
        self.filter(character=character).delete()
        skill_sets_qs = SkillSet.objects.prefetch_related(
            "skills", "skills__eve_type"
        ).all()
        skill_sets_count = skill_sets_qs.count()
        if skill_sets_count == 0:
            logger.info("%s: No skill sets defined", character)
            return UpdateSectionResult(is_changed=None, is_updated=True)

        logger.info("%s: Checking %s skill sets", character, skill_sets_count)
        skill_set_checks = [
            self.model(character=character, skill_set=skill_set)
            for skill_set in skill_sets_qs
        ]
        self.bulk_create(
            skill_set_checks, batch_size=MEMBERAUDIT_BULK_METHODS_BATCH_SIZE
        )

        # add failed recommended / required skills to objects if any
        obj_pks = list(self.filter(character=character).values_list("pk", flat=True))
        skill_set_checks = self.in_bulk(obj_pks)
        checks_by_skill_set_id = {
            obj.skill_set_id: obj for obj in skill_set_checks.values()
        }

        # required skills
        for skill_set in skill_sets_qs:
            failed_skills = self._identify_failed_skills(
                skill_set, character_skills, "required"
            )
            if failed_skills:
                checks_by_skill_set_id[skill_set.id].failed_required_skills.add(
                    *failed_skills
                )

        # required skills
        for skill_set in skill_sets_qs:
            failed_skills = self._identify_failed_skills(
                skill_set, character_skills, "recommended"
            )
            if failed_skills:
                checks_by_skill_set_id[skill_set.id].failed_recommended_skills.add(
                    *failed_skills
                )

        return UpdateSectionResult(is_changed=None, is_updated=True)

    @staticmethod
    def _identify_failed_skills(
        skill_set, character_skills: dict, level_name: str
    ) -> list:
        failed_skills = []
        kwargs = {f"{level_name}_level__isnull": False}
        for skill in skill_set.skills.filter(**kwargs):
            eve_type_id = skill.eve_type_id
            if eve_type_id not in character_skills or character_skills[
                eve_type_id
            ] < getattr(skill, f"{level_name}_level"):
                failed_skills.append(skill)

        return failed_skills


class CharacterStandingManager(GenericUpdateSimpleObjMixin, models.Manager):
    def update_or_create_esi(
        self, character: Character, force_update: bool = False
    ) -> UpdateSectionResult:
        """Update or create standing for a character from ESI."""

        return character.update_section_if_changed(
            section=character.UpdateSection.STANDINGS,
            fetch_func=self._fetch_data_from_esi,
            store_func=self._update_or_create_objs,
            force_update=force_update,
        )

    @fetch_token_for_character("esi-characters.read_standings.v1")
    def _fetch_data_from_esi(self, character: Character, token: Token) -> List[dict]:
        logger.info("%s: Fetching character standings from ESI", character)
        standings = esi.client.Character.get_characters_character_id_standings(
            character_id=character.eve_character.character_id,
            token=token.valid_access_token(),
        ).results()

        return standings

    def _update_or_create_objs(
        self, character: Character, esi_data: List[dict]
    ) -> Set[int]:
        def make_obj_from_esi_entry(character, key, value):
            obj = self.model(
                character=character,
                eve_entity=EveEntity.objects.get_or_create(id=key)[0],
                standing=value,
            )
            return obj

        return self._update_or_create_objs_generic(
            character,
            esi_data,
            esi_fields=("from_id", "standing"),
            model_fields=("eve_entity_id", "standing"),
            make_obj_from_esi_entry=make_obj_from_esi_entry,
        )


class CharacterTitleManager(GenericUpdateSimpleObjMixin, models.Manager):
    def update_or_create_esi(
        self, character: Character, force_update: bool = False
    ) -> UpdateSectionResult:
        """Update or create titles for a character from ESI."""

        return character.update_section_if_changed(
            section=character.UpdateSection.TITLES,
            fetch_func=self._fetch_data_from_esi,
            store_func=self._update_or_create_objs,
            force_update=force_update,
        )

    @fetch_token_for_character("esi-characters.read_titles.v1")
    def _fetch_data_from_esi(self, character: Character, token: Token) -> dict:
        """Update the character's roles"""

        logger.info("%s: Fetching titles from ESI", character)
        titles_data = esi.client.Character.get_characters_character_id_titles(
            character_id=character.eve_character.character_id,
            token=token.valid_access_token(),
        ).results()
        return titles_data

    def _update_or_create_objs(
        self, character: Character, esi_data: List[dict]
    ) -> Set[int]:
        def make_obj_from_esi_entry(character, key, value):
            name_sanitized = strip_tags(value).strip()[:100]
            obj = self.model(character=character, title_id=key, name=name_sanitized)
            return obj

        return self._update_or_create_objs_generic(
            character,
            esi_data,
            esi_fields=("title_id", "name"),
            model_fields=("title_id", "name"),
            make_obj_from_esi_entry=make_obj_from_esi_entry,
        )


class CharacterWalletBalanceManager(models.Manager):
    def update_or_create_esi(
        self, character: Character, force_update: bool = False
    ) -> UpdateSectionResult:
        """Update or create wallet balance for a character from ESI."""
        return character.update_section_if_changed(
            section=character.UpdateSection.WALLET_BALLANCE,
            fetch_func=self._fetch_data_from_esi,
            store_func=self._update_or_create_objs,
            force_update=force_update,
        )

    @fetch_token_for_character("esi-wallet.read_character_wallet.v1")
    def _fetch_data_from_esi(self, character: Character, token):
        logger.info("%s: Fetching wallet balance from ESI", character)
        balance = esi.client.Wallet.get_characters_character_id_wallet(
            character_id=character.eve_character.character_id,
            token=token.valid_access_token(),
        ).results()
        return balance

    def _update_or_create_objs(self, character: Character, balance):
        self.update_or_create(character=character, defaults={"total": balance})


class CharacterWalletJournalEntryManager(models.Manager):
    def update_or_create_esi(
        self, character: Character, force_update: bool = False
    ) -> UpdateSectionResult:
        """Update or create wallet journal entries for character from ESI.

        Note: Does not update unknown EveEntities.
        """
        return character.update_section_if_changed(
            section=character.UpdateSection.WALLET_JOURNAL,
            fetch_func=self._fetch_data_from_esi,
            store_func=self._update_or_create_objs,
            force_update=force_update,
        )

    @fetch_token_for_character("esi-wallet.read_character_wallet.v1")
    def _fetch_data_from_esi(self, character: Character, token: Token):
        logger.info("%s: Fetching wallet journal from ESI", character)
        journal = esi.client.Wallet.get_characters_character_id_wallet_journal(
            character_id=character.eve_character.character_id,
            token=token.valid_access_token(),
        ).results()
        return journal

    def _update_or_create_objs(self, character: Character, journal):
        cutoff_datetime = data_retention_cutoff()
        entries_list = {
            obj.get("id"): obj
            for obj in journal
            if cutoff_datetime is None or obj.get("date") > cutoff_datetime
        }
        if cutoff_datetime:
            self.filter(character=character, date__lt=cutoff_datetime).delete()

        with transaction.atomic():
            incoming_ids = set(entries_list.keys())
            existing_ids = set(self.values_list("entry_id", flat=True))
            create_ids = incoming_ids.difference(existing_ids)
            if not create_ids:
                logger.info("%s: No new wallet journal entries", character)
                return set()

            logger.info(
                "%s: Adding %s new wallet journal entries", character, len(create_ids)
            )
            entries = [
                self.model(
                    character=character,
                    entry_id=entry_id,
                    amount=row.get("amount"),
                    balance=row.get("balance"),
                    context_id=row.get("context_id"),
                    context_id_type=(
                        self.model.match_context_type_id(row.get("context_id_type"))
                    ),
                    date=row.get("date"),
                    description=row.get("description"),
                    first_party=get_or_create_or_none("first_party_id", row, EveEntity),
                    reason=row.get("reason", ""),
                    ref_type=row.get("ref_type"),
                    second_party=get_or_create_or_none(
                        "second_party_id", row, EveEntity
                    ),
                    tax=row.get("tax"),
                    tax_receiver=row.get("tax_receiver"),
                )
                for entry_id, row in entries_list.items()
                if entry_id in create_ids
            ]
            self.bulk_create(entries, batch_size=MEMBERAUDIT_BULK_METHODS_BATCH_SIZE)

        return eve_entity_ids_from_objs(entries)


class CharacterWalletTransactionManager(models.Manager):
    def update_or_create_esi(
        self, character: Character, force_update: bool = False
    ) -> UpdateSectionResult:
        """Update or create wallet transactions for a character from ESI."""
        return character.update_section_if_changed(
            section=character.UpdateSection.WALLET_TRANSACTIONS,
            fetch_func=self._fetch_data_from_esi,
            store_func=self._update_or_create_objs,
            force_update=force_update,
        )

    @fetch_token_for_character("esi-wallet.read_character_wallet.v1")
    def _fetch_data_from_esi(self, character: Character, token):
        logger.info("%s: Fetching wallet transactions from ESI", character)
        transactions = (
            esi.client.Wallet.get_characters_character_id_wallet_transactions(
                character_id=character.eve_character.character_id,
                token=token.valid_access_token(),
            ).results()
        )
        return transactions

    @fetch_token_for_character("esi-universe.read_structures.v1")
    def _update_or_create_objs(
        self, character: Character, token: Token, transactions
    ) -> Set[int]:
        from memberaudit.models import Location

        cutoff_datetime = data_retention_cutoff()
        transaction_list = {
            obj.get("transaction_id"): obj
            for obj in transactions
            if cutoff_datetime is None or obj.get("date") > cutoff_datetime
        }
        if cutoff_datetime:
            self.filter(character=character, date__lt=cutoff_datetime).delete()

        incoming_location_ids = {
            row.get("location_id") for row in transaction_list.values()
        }
        Location.objects.create_missing_esi(incoming_location_ids, token)
        type_ids = {row.get("type_id") for row in transaction_list.values()}
        EveType.objects.bulk_get_or_create_esi(ids=list(type_ids))

        eve_entity_ids = self._bulk_update_or_create(character, transaction_list)
        return eve_entity_ids

    @transaction.atomic()
    def _bulk_update_or_create(
        self, character: Character, transaction_list
    ) -> Set[int]:
        from memberaudit.models import Location

        incoming_ids = set(transaction_list.keys())
        existing_ids = set(self.values_list("transaction_id", flat=True))
        create_ids = incoming_ids.difference(existing_ids)
        if not create_ids:
            logger.info("%s: No new wallet transactions", character)
            return set()

        logger.info(
            "%s: Adding %s new wallet transactions",
            character,
            len(create_ids),
        )
        entries = []
        for transaction_id, row in transaction_list.items():
            if transaction_id in create_ids:
                try:
                    journal_entry = character.wallet_journal.get(
                        entry_id=row.get("journal_ref_id")
                    )
                except character.wallet_journal.model.DoesNotExist:
                    journal_entry = None
                entries.append(
                    self.model(
                        character=character,
                        transaction_id=transaction_id,
                        client=get_or_create_or_none("client_id", row, EveEntity),
                        date=row.get("date"),
                        is_buy=row.get("is_buy"),
                        is_personal=row.get("is_personal"),
                        journal_ref=journal_entry,
                        location=get_or_none("location_id", row, Location),
                        eve_type=EveType.objects.get(id=row.get("type_id")),
                        quantity=row.get("quantity"),
                        unit_price=row.get("unit_price"),
                    )
                )
        self.bulk_create(entries, batch_size=MEMBERAUDIT_BULK_METHODS_BATCH_SIZE)
        return eve_entity_ids_from_objs(entries)



"""Data export views."""

from django.contrib import messages
from django.contrib.auth.decorators import login_required, permission_required
from django.http import FileResponse, Http404, HttpResponse
from django.shortcuts import redirect, render
from django.utils.translation import gettext_lazy as _

from allianceauth.services.hooks import get_extension_logger
from app_utils.logging import LoggerAddTag

from memberaudit import __title__, tasks
from memberaudit.app_settings import MEMBERAUDIT_DATA_EXPORT_MIN_UPDATE_AGE
from memberaudit.core import data_exporters
from memberaudit.models import Character

from ._common import add_common_context

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


@login_required
@permission_required("memberaudit.exports_access")
def data_export(request) -> HttpResponse:
    """Render data export view."""
    topics = data_exporters.topics_and_export_files()
    context = {
        "page_title": _("Data Export"),
        "topics": topics,
        "character_count": Character.objects.count(),
        "minutes_until_next_update": MEMBERAUDIT_DATA_EXPORT_MIN_UPDATE_AGE,
    }
    return render(
        request, "memberaudit/data_export.html", add_common_context(request, context)
    )


@login_required
@permission_required("memberaudit.exports_access")
def download_export_file(request, topic: str) -> FileResponse:
    """Render file view for downloading an export file."""
    exporter = data_exporters.DataExporter.create_exporter(topic)
    destination = data_exporters.default_destination()
    zip_file = destination / exporter.output_basename.with_suffix(".zip")
    if not zip_file.exists():
        raise Http404(f"Could not find export file for {topic}")
    logger.info("Returning file %s for download of topic %s", zip_file, topic)
    return FileResponse(zip_file.open("rb"))


@login_required
@permission_required("memberaudit.exports_access")
def data_export_run_update(request, topic: str):
    """Render view for running data export update."""
    tasks.export_data_for_topic.delay(topic=topic, user_pk=request.user.pk)
    messages.info(
        request,
        _(
            "Data export for topic %s has been started. "
            "This can take a couple of minutes. "
            "You will get a notification once it is completed.",
        )
        % topic,
    )
    return redirect("memberaudit:data_export")
